__author__ = 'Lukas Leufen'
__date__ = '2020-11-05'

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler.data_handler_with_filter import DataHandlerFirFilterSingleStation, \
    DataHandlerFilterSingleStation, DataHandlerClimateFirFilterSingleStation
from mlair.data_handler.data_handler_with_filter import DataHandlerClimateFirFilter, DataHandlerFirFilter
from mlair.data_handler import DefaultDataHandler
from mlair import helpers
from mlair.helpers import to_list, sort_like
from mlair.configuration.defaults import DEFAULT_SAMPLING, DEFAULT_INTERPOLATION_LIMIT, DEFAULT_INTERPOLATION_METHOD
from mlair.helpers.filter import filter_width_kzf

import copy
import datetime as dt
from typing import Any
from functools import partial

import pandas as pd
import xarray as xr


class DataHandlerMixedSamplingSingleStation(DataHandlerSingleStation):

    def __init__(self, *args, **kwargs):
        """
        This data handler requires the kwargs sampling, interpolation_limit, and interpolation_method to be a 2D tuple
        for input and target data. If one of these kwargs is only a single argument, it will be applied to inputs and
        targets with this value. If one of these kwargs is a 2-dim tuple, the first element is applied to inputs and the
        second to targets respectively. If one of these kwargs is not provided, it is filled up with the same default
        value for inputs and targets.
        """
        self.update_kwargs("sampling", DEFAULT_SAMPLING, kwargs)
        self.update_kwargs("interpolation_limit", DEFAULT_INTERPOLATION_LIMIT, kwargs)
        self.update_kwargs("interpolation_method", DEFAULT_INTERPOLATION_METHOD, kwargs)
        super().__init__(*args, **kwargs)

    @staticmethod
    def update_kwargs(parameter_name: str, default: Any, kwargs: dict):
        """
        Update a single element of kwargs inplace to be usable for inputs and targets.

        The updated value in the kwargs dictionary is a tuple consisting on the value applicable to the inputs as first
        element and the target's value as second element: (<value_input>, <value_target>). If the value for the given
        parameter_name is already a tuple, it is checked to have exact two entries. If the paramter_name is not
        included in kwargs, the given default value is used and applied to both elements of the update tuple.

        :param parameter_name: name of the parameter that should be transformed to 2-dim
        :param default: the default value to fill if parameter is not in kwargs
        :param kwargs: the kwargs dictionary containing parameters
        """
        parameter = kwargs.get(parameter_name, default)
        if not isinstance(parameter, tuple):
            parameter = (parameter, parameter)
        assert len(parameter) == 2  # (inputs, targets)
        kwargs.update({parameter_name: parameter})

    def make_input_target(self):
        self._data = tuple(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:
        vars = [self.variables, self.target_var]
        stats_per_var = helpers.select_from_dict(self.statistics_per_var, vars[ind])
        data, self.meta = self.load_data(self.path[ind], self.station, stats_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         self.start, self.end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method[ind],
                                limit=self.interpolation_limit[ind], sampling=self.sampling[ind])

        return data

    def set_inputs_and_targets(self):
        self.input_data = self._data[0].sel({self.target_dim: helpers.to_list(self.variables)})
        self.target_data = self._data[1].sel({self.target_dim: helpers.to_list(self.target_var)})

    def setup_data_path(self, data_path, sampling):
        """Sets two paths instead of single path. Expects sampling arg to be a list with two entries"""
        assert len(sampling) == 2
        return list(map(lambda x: super(__class__, self).setup_data_path(data_path, x), sampling))

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data = f_prep(_data[0]), f_prep(_data[1])
        self.input_data, self.target_data = list(map(f_prep, [_input_data, _target_data]))


class DataHandlerMixedSampling(DefaultDataHandler):
    """Data handler using mixed sampling for input and target."""

    data_handler = DataHandlerMixedSamplingSingleStation
    data_handler_transformation = DataHandlerMixedSamplingSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithFilterSingleStation(DataHandlerMixedSamplingSingleStation,
                                                      DataHandlerFilterSingleStation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_sampling(self, **kwargs):
        assert kwargs.get("sampling") == ("hourly", "daily")

    def apply_filter(self):
        raise NotImplementedError

    def create_filter_index(self) -> pd.Index:
        """Create name for filter dimension."""
        raise NotImplementedError

    def _create_lazy_data(self):
        raise NotImplementedError

    def make_input_target(self):
        """
        A FIR filter is applied on the input data that has hourly resolution. Lables Y are provided as aggregated values
        with daily resolution.
        """
        self._data = tuple(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()
        self.apply_filter()

    def estimate_filter_width(self):
        """Return maximum filter width."""
        raise NotImplementedError

    @staticmethod
    def _add_time_delta(date, delta):
        new_date = dt.datetime.strptime(date, "%Y-%m-%d") + dt.timedelta(hours=delta)
        return new_date.strftime("%Y-%m-%d")

    def update_start_end(self, ind):
        if ind == 0:  # for inputs
            estimated_filter_width = self.estimate_filter_width()
            start = self._add_time_delta(self.start, -estimated_filter_width)
            end = self._add_time_delta(self.end, estimated_filter_width)
        else:  # target
            start, end = self.start, self.end
        return start, end

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:

        start, end = self.update_start_end(ind)
        vars = [self.variables, self.target_var]
        stats_per_var = helpers.select_from_dict(self.statistics_per_var, vars[ind])

        data, self.meta = self.load_data(self.path[ind], self.station, stats_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         start, end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method[ind],
                                limit=self.interpolation_limit[ind], sampling=self.sampling[ind])
        return data

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        start_inp, end_inp = self.update_start_end(0)
        self._data = tuple(map(lambda x: self._slice_prep(_data[x], *self.update_start_end(x)), [0, 1]))
        self.input_data = self._slice_prep(_input_data, start_inp, end_inp)
        self.target_data = self._slice_prep(_target_data, self.start, self.end)


class DataHandlerMixedSamplingWithFirFilterSingleStation(DataHandlerMixedSamplingWithFilterSingleStation,
                                                         DataHandlerFirFilterSingleStation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def estimate_filter_width(self):
        """Filter width is determined by the filter with the highest order."""
        if isinstance(self.filter_order[0], tuple):
            return max([filter_width_kzf(*e) for e in self.filter_order])
        else:
            return max(self.filter_order)

    def apply_filter(self):
        DataHandlerFirFilterSingleStation.apply_filter(self)

    def create_filter_index(self, add_unfiltered_index=True) -> pd.Index:
        return DataHandlerFirFilterSingleStation.create_filter_index(self, add_unfiltered_index=add_unfiltered_index)

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.fir_coeff, self.filter_dim_order = lazy_data
        DataHandlerMixedSamplingWithFilterSingleStation._extract_lazy(self, (_data, _meta, _input_data, _target_data))

    def _create_lazy_data(self):
        return DataHandlerFirFilterSingleStation._create_lazy_data(self)

    @staticmethod
    def _get_fs(**kwargs):
        """Return frequency in 1/day (not Hz)"""
        sampling = kwargs.get("sampling")[0]
        if sampling == "daily":
            return 1
        elif sampling == "hourly":
            return 24
        else:
            raise ValueError(f"Unknown sampling rate {sampling}. Only daily and hourly resolution is supported.")


class DataHandlerMixedSamplingWithFirFilter(DataHandlerFirFilter):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithFirFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithFirFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithClimateFirFilterSingleStation(DataHandlerClimateFirFilterSingleStation,
                                                                DataHandlerMixedSamplingWithFirFilterSingleStation):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.climate_filter_coeff, self.apriori, self.all_apriori, \
        self.filter_dim_order = lazy_data
        DataHandlerMixedSamplingWithFilterSingleStation._extract_lazy(self, (_data, _meta, _input_data, _target_data))


class DataHandlerMixedSamplingWithClimateFirFilter(DataHandlerClimateFirFilter):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    data_handler_unfiltered = DataHandlerMixedSamplingSingleStation
    _requirements = list(set(data_handler.requirements() + data_handler_unfiltered.requirements()))
    DEFAULT_FILTER_ADD_UNFILTERED = False

    def __init__(self, *args, data_handler_class_unfiltered: data_handler_unfiltered = None,
                 filter_add_unfiltered: bool = DEFAULT_FILTER_ADD_UNFILTERED, **kwargs):
        self.dh_unfiltered = data_handler_class_unfiltered
        self.filter_add_unfiltered = filter_add_unfiltered
        super().__init__(*args, **kwargs)

    def _create_collection(self):
        collection = super()._create_collection()
        if self.filter_add_unfiltered is True and self.dh_unfiltered is not None:
            collection.append(self.dh_unfiltered)
        return collection

    @classmethod
    def build(cls, station: str, **kwargs):
        sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler.requirements() if k in kwargs}
        filter_add_unfiltered = kwargs.get("filter_add_unfiltered", False)
        sp_keys = cls.build_update_transformation(sp_keys, dh_type="filtered")
        sp = cls.data_handler(station, **sp_keys)
        if filter_add_unfiltered is True:
            sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler_unfiltered.requirements() if k in kwargs}
            sp_keys = cls.build_update_transformation(sp_keys, dh_type="unfiltered")
            sp_unfiltered = cls.data_handler_unfiltered(station, **sp_keys)
        else:
            sp_unfiltered = None
        dp_args = {k: copy.deepcopy(kwargs[k]) for k in cls.own_args("id_class") if k in kwargs}
        return cls(sp, data_handler_class_unfiltered=sp_unfiltered, **dp_args)

    @classmethod
    def build_update_transformation(cls, kwargs_dict, dh_type="filtered"):
        if "transformation" in kwargs_dict:
            trafo_opts = kwargs_dict.get("transformation")
            if isinstance(trafo_opts, dict):
                kwargs_dict["transformation"] = trafo_opts.get(dh_type)
        return kwargs_dict

    @classmethod
    def transformation(cls, set_stations, tmp_path=None, dh_transformation=None, **kwargs):

        # sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls._requirements if k in kwargs}
        if "transformation" not in kwargs.keys():
            return

        if dh_transformation is None:
            dh_transformation = (cls.data_handler_transformation, cls.data_handler_unfiltered)
        elif not isinstance(dh_transformation, tuple):
            dh_transformation = (dh_transformation, dh_transformation)
        transformation_filtered = super().transformation(set_stations, tmp_path=tmp_path,
                                                         dh_transformation=dh_transformation[0], **kwargs)
        if kwargs.get("filter_add_unfiltered", False) is False:
            return transformation_filtered
        else:
            transformation_unfiltered = super().transformation(set_stations, tmp_path=tmp_path,
                                                               dh_transformation=dh_transformation[1], **kwargs)
            return {"filtered": transformation_filtered, "unfiltered": transformation_unfiltered}

    def get_X_original(self):
        if self.use_filter_branches is True:
            X = []
            for data in self._collection:
                if hasattr(data, "filter_dim"):
                    X_total = data.get_X()
                    filter_dim = data.filter_dim
                    for filter_name in data.filter_dim_order:
                        X.append(X_total.sel({filter_dim: filter_name}, drop=True))
                else:
                    X.append(data.get_X())
            return X
        else:
            return super().get_X_original()


class DataHandlerMixedSamplingWithClimateAndFirFilter(DataHandlerMixedSamplingWithClimateFirFilter):
    # data_handler = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    # data_handler_transformation = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    # data_handler_unfiltered = DataHandlerMixedSamplingSingleStation
    # _requirements = list(set(data_handler.requirements() + data_handler_unfiltered.requirements()))
    # DEFAULT_FILTER_ADD_UNFILTERED = False
    data_handler_climate_fir = DataHandlerMixedSamplingWithClimateFirFilterSingleStation
    data_handler_fir = (DataHandlerMixedSamplingWithFirFilterSingleStation,
                        DataHandlerMixedSamplingWithClimateFirFilterSingleStation)
    data_handler_fir_pos = None
    data_handler = None
    data_handler_unfiltered = DataHandlerMixedSamplingSingleStation
    _requirements = list(set(data_handler_climate_fir.requirements() + data_handler_fir[0].requirements() +
                             data_handler_fir[1].requirements() + data_handler_unfiltered.requirements()))
    chem_indicator = "chem"
    meteo_indicator = "meteo"

    def __init__(self, data_handler_class_chem, data_handler_class_meteo, data_handler_class_chem_unfiltered,
                 data_handler_class_meteo_unfiltered, chem_vars, meteo_vars, *args, **kwargs):

        if len(chem_vars) > 0:
            id_class, id_class_unfiltered = data_handler_class_chem, data_handler_class_chem_unfiltered
            self.id_class_other = data_handler_class_meteo
            self.id_class_other_unfiltered = data_handler_class_meteo_unfiltered
        else:
            id_class, id_class_unfiltered = data_handler_class_meteo, data_handler_class_meteo_unfiltered
            self.id_class_other = data_handler_class_chem
            self.id_class_other_unfiltered = data_handler_class_chem_unfiltered
        super().__init__(id_class, *args, data_handler_class_unfiltered=id_class_unfiltered, **kwargs)

    @classmethod
    def _split_chem_and_meteo_variables(cls, **kwargs):
        if "variables" in kwargs:
            variables = kwargs.get("variables")
        elif "statistics_per_var" in kwargs:
            variables = kwargs.get("statistics_per_var")
        else:
            variables = None
        if variables is None:
            variables = cls.data_handler_climate_fir.DEFAULT_VAR_ALL_DICT.keys()
        chem_vars = cls.data_handler_climate_fir.chem_vars
        chem = set(variables).intersection(chem_vars)
        meteo = set(variables).difference(chem_vars)
        return sort_like(to_list(chem), variables), sort_like(to_list(meteo), variables)

    @classmethod
    def build(cls, station: str, **kwargs):
        chem_vars, meteo_vars = cls._split_chem_and_meteo_variables(**kwargs)
        filter_add_unfiltered = kwargs.get("filter_add_unfiltered", False)
        sp_chem, sp_chem_unfiltered = None, None
        sp_meteo, sp_meteo_unfiltered = None, None

        if len(chem_vars) > 0:
            sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler_climate_fir.requirements() if k in kwargs}
            sp_keys = cls.build_update_transformation(sp_keys, dh_type="filtered_chem")

            cls.prepare_build(sp_keys, chem_vars, cls.chem_indicator)
            sp_chem = cls.data_handler_climate_fir(station, **sp_keys)
            if filter_add_unfiltered is True:
                sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler_unfiltered.requirements() if k in kwargs}
                sp_keys = cls.build_update_transformation(sp_keys, dh_type="unfiltered_chem")
                cls.prepare_build(sp_keys, chem_vars, cls.chem_indicator)
                sp_chem_unfiltered = cls.data_handler_unfiltered(station, **sp_keys)
        if len(meteo_vars) > 0:
            if cls.data_handler_fir_pos is None:
                if "extend_length_opts" in kwargs:
                    if isinstance(kwargs["extend_length_opts"], dict) and cls.meteo_indicator not in kwargs["extend_length_opts"].keys():
                        cls.data_handler_fir_pos = 0  # use faster fir version without climate estimate
                    else:
                        cls.data_handler_fir_pos = 1  # use slower fir version with climate estimate
                else:
                    cls.data_handler_fir_pos = 0  # use faster fir version without climate estimate
            sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler_fir[cls.data_handler_fir_pos].requirements() if k in kwargs}
            sp_keys = cls.build_update_transformation(sp_keys, dh_type="filtered_meteo")
            cls.prepare_build(sp_keys, meteo_vars, cls.meteo_indicator)
            sp_meteo = cls.data_handler_fir[cls.data_handler_fir_pos](station, **sp_keys)
            if filter_add_unfiltered is True:
                sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls.data_handler_unfiltered.requirements() if k in kwargs}
                sp_keys = cls.build_update_transformation(sp_keys, dh_type="unfiltered_meteo")
                cls.prepare_build(sp_keys, meteo_vars, cls.meteo_indicator)
                sp_meteo_unfiltered = cls.data_handler_unfiltered(station, **sp_keys)

        dp_args = {k: copy.deepcopy(kwargs[k]) for k in cls.own_args("id_class") if k in kwargs}
        return cls(sp_chem, sp_meteo, sp_chem_unfiltered, sp_meteo_unfiltered, chem_vars, meteo_vars, **dp_args)

    @classmethod
    def prepare_build(cls, kwargs, var_list, var_type):
        kwargs.update({"variables": var_list})
        for k in list(kwargs.keys()):
            v = kwargs[k]
            if isinstance(v, dict):
                if len(set(v.keys()).intersection({cls.chem_indicator, cls.meteo_indicator})) > 0:
                    try:
                        new_v = kwargs.pop(k)
                        kwargs[k] = new_v[var_type]
                    except KeyError:
                        pass

    @staticmethod
    def adjust_window_opts(key: str, parameter_name: str, kwargs: dict):
        try:
            if parameter_name in kwargs:
                window_opt = kwargs.pop(parameter_name)
                if isinstance(window_opt, dict):
                    window_opt = window_opt[key]
                kwargs[parameter_name] = window_opt
        except KeyError:
            pass

    def _create_collection(self):
        collection = super()._create_collection()
        if self.id_class_other is not None:
            collection.append(self.id_class_other)
        if self.filter_add_unfiltered is True and self.id_class_other_unfiltered is not None:
            collection.append(self.id_class_other_unfiltered)
        return collection

    @classmethod
    def transformation(cls, set_stations, tmp_path=None, **kwargs):

        if "transformation" not in kwargs.keys():
            return

        chem_vars, meteo_vars = cls._split_chem_and_meteo_variables(**kwargs)
        transformation_chem, transformation_meteo = None, None
        # chem transformation
        if len(chem_vars) > 0:
            kwargs_chem = copy.deepcopy(kwargs)
            cls.prepare_build(kwargs_chem, chem_vars, cls.chem_indicator)
            dh_transformation = (cls.data_handler_climate_fir, cls.data_handler_unfiltered)
            transformation_chem = super().transformation(set_stations, tmp_path=tmp_path,
                                                         dh_transformation=dh_transformation, **kwargs_chem)

        # meteo transformation
        if len(meteo_vars) > 0:
            kwargs_meteo = copy.deepcopy(kwargs)
            cls.prepare_build(kwargs_meteo, meteo_vars, cls.meteo_indicator)
            dh_transformation = (cls.data_handler_fir[cls.data_handler_fir_pos or 0], cls.data_handler_unfiltered)
            transformation_meteo = super().transformation(set_stations, tmp_path=tmp_path,
                                                          dh_transformation=dh_transformation, **kwargs_meteo)

        # combine all transformations
        transformation_res = {}
        if transformation_chem is not None:
            if isinstance(transformation_chem, dict):
                if len(transformation_chem) > 0:
                    transformation_res["filtered_chem"] = transformation_chem.pop("filtered")
                    transformation_res["unfiltered_chem"] = transformation_chem.pop("unfiltered")
            else:  # if no unfiltered chem branch
                transformation_res["filtered_chem"] = transformation_chem
        if transformation_meteo is not None:
            if isinstance(transformation_meteo, dict):
                if len(transformation_meteo) > 0:
                    transformation_res["filtered_meteo"] = transformation_meteo.pop("filtered")
                    transformation_res["unfiltered_meteo"] = transformation_meteo.pop("unfiltered")
            else:  # if no unfiltered meteo branch
                transformation_res["filtered_meteo"] = transformation_meteo
        return transformation_res if len(transformation_res) > 0 else None

    def get_X_original(self):
        if self.use_filter_branches is True:
            X = []
            for data in self._collection:
                if hasattr(data, "filter_dim"):
                    X_total = data.get_X()
                    filter_dim = data.filter_dim
                    for filter_name in data.filter_dim_order:
                        X.append(X_total.sel({filter_dim: filter_name}, drop=True))
                else:
                    X.append(data.get_X())
            return X
        else:
            return super().get_X_original()
