__author__ = "Felix Kleinert"
__data__ = "2021-02-03"

import xarray as xr
import numpy as np
import pandas as pd
import itertools
import glob
import matplotlib.pyplot as plt
from dask.diagnostics import ProgressBar
from tzwhere import tzwhere
import datetime as dt
from toarstats import toarstats

import hashlib
from functools import reduce, partial
import dill


import dask
import inspect
import os
import gc
from mlair.helpers.geofunctions import haversine_dist, bearing_angle, WindSector, VectorRotateLambertConformal2latlon
from mlair.helpers.helpers import convert2xrda, remove_items, to_list
from mlair.helpers import TimeTrackingWrapper, TimeTracking
from mlair.configuration import check_path_and_create


from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler

from typing import Tuple, Union, Dict, List
import logging

float_np_xr = Union[float, np.ndarray, xr.DataArray, xr.Dataset]
int_or_list_of_int = Union[int, List[int]]


class BaseWrfChemDataLoader:
    """
    Base clase to load WRF-Chem data.
    """
    DEFAULT_LOGICAL_TIME_COORD_NAME = 'Time'
    DEFAULT_LOGICAL_X_COORD_NAME = 'west_east'
    DEFAULT_LOGICAL_Y_COORD_NAME = 'south_north'
    DEFAULT_LOGICAL_Z_COORD_NAME = 'bottom_top'

    DEFAULT_PHYSICAL_TIME_COORD_NAME = "XTIME"
    DEFAULT_PHYSICAL_X_COORD_NAME = 'XLONG'
    DEFAULT_PHYSICAL_Y_COORD_NAME = 'XLAT'

    DEFAULT_RECHUNK = {"Time": -1, "y": 36, "x": 40}
    DEFAULT_FILE_STARTER = 'wrfout_d0'

    DEFAULT_STAGED_VARS = ('U', 'V')
    DEFAULT_STAGED_ROTATION_opts = dict(cen_lon=12., cen_lat=52.5,
                                        truelat1=30., truelat2=60.,
                                        stand_lon=12.)
    DEFAULT_VARS_TO_ROTATE = ((('U', 'V'), ('Ull', 'Vll')),
                              (('U10', 'V10'), ('U10ll', 'V10ll')),
                              (('U', 'V'), ('wspdll', 'wdirll')),
                              (('U10', 'V10'), ('wspd10ll', 'wdir10ll'))
                              )

    def __init__(self, data_path: str, common_file_starter: str = DEFAULT_FILE_STARTER,
                 time_dim_name: str = DEFAULT_LOGICAL_TIME_COORD_NAME,
                 rechunk_values: Dict = None,
                 logical_x_coord_name: str = DEFAULT_LOGICAL_X_COORD_NAME,
                 logical_y_coord_name: str = DEFAULT_LOGICAL_Y_COORD_NAME,
                 logical_z_coord_name: str = DEFAULT_LOGICAL_Z_COORD_NAME,
                 physical_x_coord_name: str = DEFAULT_PHYSICAL_X_COORD_NAME,
                 physical_y_coord_name: str = DEFAULT_PHYSICAL_Y_COORD_NAME,
                 physical_t_coord_name: str = DEFAULT_PHYSICAL_TIME_COORD_NAME,
                 variables=None, z_coord_selector=None,
                 start_time=None, end_time=None,
                 staged_rotation_opts: Dict = DEFAULT_STAGED_ROTATION_opts,
                 vars_to_rotate: Tuple[Tuple[Tuple[str, str], Tuple[str, str]]] = DEFAULT_VARS_TO_ROTATE,
                 staged_dimension_mapping=None, stag_ending='_stag',
                 date_format_of_nc_file=None, vars_for_unit_conv: Dict = None
                 ):
        """
        Initialisze data loader

        :param data_path: path to WRF-Chem data
        :param common_file_starter: beginning of each idividual file
        :param time_dim_name: Name of time diension in data file
        :param rechunk_values: Mapping for chunk sizes
        :param logical_x_coord_name: Logical x coord name (e.g. west-east direction)
        :param logical_y_coord_name: Logical y coord name (e.g. south-north direction)
        :param logical_z_coord_name: Logical z coord name (e.g. bottom-top of atmosphere direction)
        :param physical_x_coord_name: Physical x coord name (e.g. XLONG, might be 2D)
        :param physical_y_coord_name: Physical y coord name (e.g. XLAT, might be 2D)
        :param physical_t_coord_name: Physical time coord name
        :param variables: variables to read from files
        :param z_coord_selector: physical z coords to read from file
        :param start_time: first physical time coord to read from file
        :param end_time: end physical time coord to read from file
        :param staged_rotation_opts: mapping of options needed for vector transformation
        :param vars_to_rotate: name of variables which should be rotated from one coord system to another
                (e.g.  ((('U', 'V'), ('Ull', 'Vll')), (('U10', 'V10'), ('U10ll', 'V10ll'))) to rotate U and V
                in coord sys A to Ull and Vll in coord sys B (e.g. geogr)).
        :param staged_dimension_mapping: mapping of staged dimensions and names
        :type staged_dimension_mapping:
        :param stag_ending: ending of staged variables identifier (most likely provided in the metadata of the WRF-Chem
        file)
        :param date_format_of_nc_file: date format of input  file names (e.g %Y-%m-%d)
        """
        # super().__init__()
        self.data_path = data_path
        self.common_file_starter = common_file_starter
        self.time_dim_name = time_dim_name

        self.logical_x_coord_name = logical_x_coord_name
        self.logical_y_coord_name = logical_y_coord_name
        self.logical_z_coord_name = logical_z_coord_name

        self.physical_x_coord_name = physical_x_coord_name
        self.physical_y_coord_name = physical_y_coord_name
        self.physical_t_coord_name = physical_t_coord_name

        self.staged_rotation_opts = staged_rotation_opts
        self.vars_to_rotate = vars_to_rotate

        self.variables = variables
        self.z_coord_selector = z_coord_selector

        self.start_time = start_time
        self.end_time = end_time
        if date_format_of_nc_file is None:
            date_format_of_nc_file = "%Y-%m-%d"
        self.date_format_of_nc_file = date_format_of_nc_file

        self.rechunk_values = rechunk_values

        self._stag_ending = stag_ending
        if staged_dimension_mapping is None:
            self.staged_dimension_mapping = {'X': self.logical_x_coord_name+self._stag_ending,
                                             'Y': self.logical_y_coord_name+self._stag_ending,
                                             'Z': self.logical_z_coord_name+self._stag_ending,}
        else:
            self.staged_dimension_mapping = staged_dimension_mapping

        self.vars_for_unit_conv = vars_for_unit_conv
        # chemical convs.
        self._parts_per_exponents = {'ppmv': 6, 'ppbv': 9, 'pptv': 12, 'ppqv': 15}

        # internal
        self._X = None
        self._Y = None
        self._data = None

    @property
    def staged_dimension_mapping(self):
        return self.__staged_dimension_mapping

    @staged_dimension_mapping.setter
    def staged_dimension_mapping(self, var: dict):
        if not isinstance(var, dict):
            raise TypeError(f"`var' must be a dict, but is of type {type(var)}")
        self.__staged_dimension_mapping = var

    @property
    def data(self) -> xr.Dataset:
        """
        Returns the actual data of the WRF-Chem files
        :return:
        :rtype:
        """
        return self._data

    @property
    def dataset_search_str(self):
        """
        Returns a string to search for existing WRF-Chem data
        :return:
        :rtype:
        """
        if (self.start_time is None) and (self.end_time is None):
            path_list = os.path.join(self.data_path, self.common_file_starter + '*')
            #logging.info(f"Reading file(s): {path_list}")
            return path_list
        elif (self.start_time is not None) and (self.end_time is not None):
            path_list = set()
            for day in pd.date_range(self.start_time, self.end_time):
                path_list.add(sorted(glob.glob(os.path.join(self.data_path,
                                                            self.common_file_starter) + '*' + day.strftime(
                    self.date_format_of_nc_file) + '*'
                                               ))[0])
            path_list = sorted(list(path_list))
            #logging.info(f"Reading file(s): {path_list}")
            return path_list
        else:
            raise ValueError(f"`start_time' and `end_time' must both be given or None.")

    def convert_chem(self, data, from_unit, to_unit):
        convert_exponent = self._parts_per_exponents[to_unit] - self._parts_per_exponents[from_unit]
        return data * 10**convert_exponent

    @TimeTrackingWrapper
    def open_data(self):
        """
        Opens WRF-Chem data as dataset
        :return:
        :rtype:
        """
        if self.variables is None:
            # see also https://github.com/pydata/xarray/issues/1385#issuecomment-438870575
            data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,
                                     parallel=True, decode_cf=False)
            data = xr.decode_cf(data)
            logging.info(f"Reading file(s): {self.dataset_search_str}")
        else:
            data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,

                                     parallel=True, preprocess=self.preprocess_fkt_for_loader,
                                     #decode_cf=False,
                                     )
            logging.info(f"Reading file(s): {self.dataset_search_str}")
        #data = xr.decode_cf(data)
        self._data = data

    def preprocess_fkt_for_loader(self, ds):
        # ToDo make genreal: Currently it's in fact hardcoded!
        potential_wind_vars_list = list(
            set(itertools.chain(
                *itertools.chain(*self.vars_to_rotate))))
        none_wind_vars_to_keep = [x for x in self.variables if x not in potential_wind_vars_list]
        wind_vars_to_keep = ['U', 'V', 'U10', 'V10']
        combined_vars_to_keep = none_wind_vars_to_keep + wind_vars_to_keep
        ds = ds[combined_vars_to_keep]
        if self.z_coord_selector is not None:
            ds = ds.sel({self.logical_z_coord_name: self.z_coord_selector})
        return ds

    def assign_coords(self, coords, **coords_kwargs):
        """
        Assign coords to WrfChemDataHandler._data

        """
        self._data = self._data.assign_coords(coords, **coords_kwargs)

    def rechunk_data(self, chunks=None, name_prefix='xarray-', token=None, lock=False):
        return self._data.chunk(chunks=chunks, name_prefix=name_prefix, token=token, lock=lock)

    def read_filenames(self):
        combs = []
        start = 1
        end = 9
        for l in range(1, end):
            combs.extend(itertools.combinations(range(start, end), l))

        for i in combs:
            filenamestr = f'{self.dataset_search_str[:-1]}{list(i)}_*'
            try:
                with xr.open_mfdataset(paths=filenamestr, combine='nested', concat_dim='Time',
                                       parallel=True, chunks={'Time': 12}) as data:
                    status = 'OK'
            except:
                status = 'FAIL'
            print(f'{i}: {filenamestr} {status}')

    def get_distances(self, lat, lon):
        """
        Calculates distances from given lat, lon to all grid boxes of model domain.
        :param lat: latitude
        :param lon: longitude
        :return:
        :rtype:
        """
        dist = haversine_dist(lat1=self._data.XLAT, lon1=self._data.XLONG, lat2=lat, lon2=lon)
        return dist

    def get_bearing(self, lat, lon, points_last=True):
        """
        Calculates bearings from a given lat, lon to all grid boxes of model domain.
        :param lat: latitude
        :param lon:  longitude
        :param points_last: return transpose with points as last dimensions (if True, default)
        :return:
        :rtype:
        """
        bearing = bearing_angle(lat1=lat, lon1=lon, lat2=self._data.XLAT, lon2=self._data.XLONG)
        if points_last:
            return bearing.transpose(..., 'points')
        else:
            return bearing

    def compute_nearest_icoordinates(self, lat, lon, dim=None):
        """
        Calculate nearest logical coordinate (grid box) for a given location (lat, lon)

        :param lat: latitude
        :param lon: longitude
        :param dim: dimension to apply argmin function
        :return:
        :rtype:
        """
        dist = self.get_distances(lat=lat, lon=lon)

        if dim is None:
            return dist.argmin(...)
        else:
            return dist.argmin(dim)

    def _set_dims_as_coords(self):
        """
        Set dimensions as coordinates
        :return:
        :rtype:
        """
        if self._data is None:
            raise IOError(f'{self.__class__.__name__} can not set dims as coords. Use must use `open_data()` before.')
        data = self._data
        for k, _ in data.dims.items():
            data = data.assign_coords({k: data[k]})
        self._data = data
        # logging.info('set dimensions as coordinates')

    def apply_staged_transormation(self, mapping_of_stag2unstag=None):
        """
        Apply vector transformation on staged variables

        :param mapping_of_stag2unstag:mapping from staged field as key to unstaged field as value. Note that U10 and V10
        in WRF-Chem are located at grid box centers while U and V are located on edges.
        :type mapping_of_stag2unstag:
        :return:
        :rtype:
        """
        if mapping_of_stag2unstag is None:
            mapping_of_stag2unstag = {'U': 'U10', 'V': 'V10', 'U10': 'U10', 'V10': 'V10'}
        vectorrot = VectorRotateLambertConformal2latlon(xlat=self._data[self.physical_y_coord_name].data,
                                                        xlong=self._data[self.physical_x_coord_name].data,
                                                        **self.staged_rotation_opts)
        if self._data is None:
            raise IOError(f"`open_data' must be called before vector interpolation and rotation")
        for (u_stag_name, v_stag_name), (u_ll_name, v_ll_name) in self.vars_to_rotate:
            print((u_stag_name, v_stag_name), (u_ll_name, v_ll_name))
            try:
                u_staged_field = self._data[u_stag_name]
                v_staged_field = self._data[v_stag_name]

                u_target_field = self._data[mapping_of_stag2unstag[u_stag_name]]
                v_target_field = self._data[mapping_of_stag2unstag[v_stag_name]]

                # This if statements are needed as cdo might have removed empty attrs from variables
                if "stagger" not in u_staged_field.attrs:
                    u_staged_field.attrs['stagger'] = ""
                if "stagger" not in v_staged_field.attrs:
                    v_staged_field.attrs['stagger'] = ""

                u_stagger = u_staged_field.attrs['stagger']
                v_stagger = v_staged_field.attrs['stagger']

                if u_stagger:
                    u_grd = self.set_interpolated_field(staged_field=u_staged_field, target_field=u_target_field)
                else:
                    u_grd = self._data[u_stag_name]

                if v_stagger:
                    v_grd = self.set_interpolated_field(staged_field=v_staged_field, target_field=v_target_field)
                else:
                    v_grd = self._data[v_stag_name]

                if v_grd.dims != (self.physical_t_coord_name, self.logical_y_coord_name, self.logical_x_coord_name):
                    v_grd = v_grd.transpose(self.physical_t_coord_name, self.logical_y_coord_name,
                                            self.logical_x_coord_name)
                if u_grd.dims != (self.physical_t_coord_name, self.logical_y_coord_name, self.logical_x_coord_name):
                    u_grd = u_grd.transpose(self.physical_t_coord_name, self.logical_y_coord_name,
                                            self.logical_x_coord_name)

                if u_ll_name[0] == 'U' and v_ll_name[0] == 'V':
                    ull, vll = vectorrot.ugrd_vgrd2ull_vll(u_grd, v_grd)
                elif u_ll_name[:4] == 'wspd' and v_ll_name[:4] == 'wdir':
                    ull, vll = vectorrot.ugrd_vgrd2wspd_wdir(u_grd, v_grd)
                else:
                    raise ValueError(
                        f"`u_ll_name' and `u_ll_name' must either start with 'U' and 'V' or 'wspd' and 'wdir', "
                        f"but they are u_ll_name={u_ll_name} and v_ll_name={v_ll_name}")

                self._data[u_ll_name] = ull
                self._data[v_ll_name] = vll
            except KeyError:
                pass

    def set_interpolated_field(self, staged_field: xr.DataArray, target_field: xr.DataArray,
                               dropped_staged_attrs: List[str] =None, **kwargs):
        """
        Interpolates a staged field (variables on edges of grid box) to a given target field (center of grid box)

        :param staged_field: field of staged variables
        :param target_field: target field to interpolate on
        :param dropped_staged_attrs: attribute which indicates that field is staged and therefore should be removed
        as attribute for the target field
        :param kwargs:
        :return: Interpolated field
        :rtype:
        """
        stagger_attr_name = kwargs.pop('stagger_attr_name', 'stagger')
        stagger = kwargs.pop('stagger', staged_field.attrs[stagger_attr_name])
        if dropped_staged_attrs is None:
            dropped_staged_attrs = [stagger_attr_name]

        # prepare new (interpolated) field by setting metainfo etc
        excluded_dims = [self.staged_dimension_mapping[stagger]]
        target_dim_order = [dim.replace('_stag', '') for dim in staged_field.dims]
        new_field = xr.broadcast(target_field, staged_field, exclude=excluded_dims)[0]
        new_field = new_field.transpose(*target_dim_order)
        new_field.attrs.update(remove_items(staged_field.attrs, dropped_staged_attrs))

        # interpolate from staged to unstaged grid
        new_field.data = VectorRotateLambertConformal2latlon.interpolate_to_grid_center(staged_field,
                                                                                        self.staged_dimension_mapping[stagger])
        return new_field


class SingleGridColumnWrfChemDataLoader(BaseWrfChemDataLoader):
    DEFAULT_MODEL = "WRF-Chem"
    DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                            'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                            'pblheight': 'maximum'}
    DEFAULT_WINDOW_LEAD_TIME = 3
    DEFAULT_WINDOW_HISTORY_SIZE = 13
    DEFAULT_WINDOW_HISTORY_OFFSET = 0
    DEFAULT_TARGET_VAR = "o3"
    DEFAULT_TARGET_DIM = "variables"
    DEFAULT_ITER_DIM = "points"
    DEFAULT_WINDOW_DIM = "window"

    _hash = ["data_path", "external_coords_file", "time_dim_name",
             "rechunk_values", "variables", "z_coord_selector", "date_format_of_nc_file",
             "wind_sectors", "wind_sector_edge_dim_name", "statistics_per_var",
             "aggregation_sampling", "time_zone",
             ]

    def __init__(self, coords: Tuple[float_np_xr, float_np_xr],
                 target_dim=DEFAULT_TARGET_DIM, target_var=DEFAULT_TARGET_VAR,
                 window_history_size=DEFAULT_WINDOW_HISTORY_SIZE,
                 window_lead_time=DEFAULT_WINDOW_LEAD_TIME,
                 external_coords_file: str = None,
                 wind_sectors=None, wind_sector_edge_dim_name=None, statistics_per_var: Dict = None,
                 aggregation_sampling: str = None, time_zone: str = None, target_time_type: str = None,
                 station: str = None, lazy_preprocessing: bool = False, **kwargs):
        """


        :param coords:
        :type coords:
        :param target_dim:
        :type target_dim:
        :param target_var:
        :type target_var:
        :param window_history_size:
        :type window_history_size:
        :param window_lead_time:
        :type window_lead_time:
        :param external_coords_file:
        :type external_coords_file:
        :param wind_sectors:
        :type wind_sectors:
        :param statistics_per_var: Dict containing the (TOAR-) statistics as value for each variable (key)
        :param aggregation_sampling: sampling period for statistics (e.g. "daily", "seasonal" etc. See torstats for more details)
        :param time_zone:
        :param: station
        :param kwargs:
        :type kwargs:
        """
        super().__init__(**kwargs)
        self._set_coords(coords)
        self.target_dim = target_dim
        self.target_var = target_var
        self.window_history_size = window_history_size
        self.window_lead_time = window_lead_time
        self._nearest_icoords = None
        self._nearest_coords = None
        self.external_coords_file = external_coords_file
        self.wind_sectors = wind_sectors
        self.wind_sector_edge_dim_name = wind_sector_edge_dim_name
        self.statistics_per_var = statistics_per_var
        self.aggregation_sampling = aggregation_sampling
        self.time_zone = time_zone
        self.target_time_type = target_time_type
        self.station = station

        self.lazy = lazy_preprocessing
        self.lazy_path = None
        if self.lazy is True:
            self.lazy_path = os.path.join(self.data_path, "lazy_data", self.__class__.__name__)
            check_path_and_create(self.lazy_path)


        logging.debug("SingleGridColumnWrfChemDataLoader Initialised")

    def __enter__(self):
        """
        Enter method which opens WRF-Chem data, sets coordinates (potentially from external coord file)

        :return:
        :rtype:
        """
        self.open_data()

        if self.physical_t_coord_name != self.time_dim_name:
            self.assign_coords(
                {self.physical_t_coord_name: (self.time_dim_name, self._data[self.physical_t_coord_name].values)})

        self._set_dims_as_coords()
        if self.external_coords_file is not None:
            self._apply_external_coordinates()

        # self.apply_staged_transormation()
        # self._set_geoinfos()

        if self.vars_for_unit_conv is not None:
            self.convert_chemical_units()

        if self.lazy is False:
            self.reset_data_by_other(self.apply_toarstats())
            self.apply_staged_transormation()
        else:
            self.load_lazy()
            self.store_lazy()


        self._set_geoinfos()
        # self.rechunk_data(self.rechunk_values)
        return self

    def get_wrfnames_from_statistics_per_var(self):
        """
        Returns U and V or U10 and V10 instead of winddir windspd due to averaging.
        :return:
        :rtype:
        """
        working_dict = self.statistics_per_var.copy()
        replacer = ((("U10", "V10"), ("wdir10ll", "wspd10ll")),
                    (("U", "V"), ("wdirll", "wspdll"))
                    )
        for (u_rep, v_rep), (wdir_org, wspd_org) in replacer:
            if wdir_org in self.statistics_per_var.keys():
            # try:
                working_dict[u_rep] = self.statistics_per_var[wdir_org]
                working_dict[v_rep] = self.statistics_per_var[wdir_org]
                working_dict.pop(wdir_org)
                if wspd_org in self.statistics_per_var.keys():
                # try:
                    working_dict.pop(wspd_org)
                # except:
                #     pass
            # except:
            #     pass

        return working_dict

    def reset_data_by_other(self, other: xr.Dataset):
        attrs = self._data.attrs
        self._data = other
        # for var in other:
        #     self._data[var] = other[var]

    def store_lazy(self):
        hash = self._get_hash()
        filename = os.path.join(self.lazy_path, hash + ".nc")
        if not os.path.exists(filename):
            self._data.to_netcdf(filename, format="NETCDF4", engine="netcdf4")
            logging.info(f"{self.station[0]}: store lazy data: {filename}")
            # dill.dump(self._create_lazy_data(), file=open(filename, "wb"))

    # def _create_lazy_data(self):
    #     return [self._data ]#, self.meta, self.input_data, self.target_data]

    def load_lazy(self):
        hash = self._get_hash()
        filename = os.path.join(self.lazy_path, hash + ".nc")
        try:
            self._data = xr.open_mfdataset(filename)
            logging.info(f"{self.station[0]}: used lazy data: {filename}")
        except FileNotFoundError:
            logging.debug(f"{self.station[0]}: could not use lazy data")
            self.reset_data_by_other(self.apply_toarstats())
            self.apply_staged_transormation()
        except OSError:
            logging.debug(f"{self.station[0]}: could not use lazy data")
            self.reset_data_by_other(self.apply_toarstats())
            self.apply_staged_transormation()

    def _extract_lazy(self, lazy_data):
        # _data, self.meta, _input_data, _target_data = lazy_data
        # f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data = lazy_data[0]
        # self._data, self.input_data, self.target_data = list(map(f_prep, [_data, _input_data, _target_data]))

    def _hash_list(self):
        return sorted(list(set(self._hash)))

    def _get_hash(self):
        hash = "".join([str(self.__getattribute__(e)) for e in self._hash_list()]).encode()
        return hashlib.md5(hash).hexdigest()

    def convert_chemical_units(self):
        with xr.set_options(keep_attrs=True):
            for var, to_unit in self.vars_for_unit_conv.items():
                from_unit = self.data[var].attrs['units']
                data = self.convert_chem(self.data[var], from_unit, to_unit)
                data.attrs['units'] = to_unit
                self.data[var] = data

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.data.close()
        gc.collect()

    def _set_geoinfos(self):
        """
        Sets geoinfos like nearest logical and physical coordinates, distances and bearings to/from a point of interest.

        :return:
        :rtype:
        """
        # identify nearest coords
        self._set_nearest_icoords(dim=[self.logical_x_coord_name, self.logical_y_coord_name])
        self._set_nearest_coords()

        # get lat and lons
        lat = self.get_coordinates()['lat']
        lon = self.get_coordinates()['lon']

        # calculate distances to points (from points to all grid boxes in lat/lon space)
        dist = self.get_distances(lat, lon)
        dist.attrs.update(dict(units='km'))

        # calculate bearing angle
        bearing = self.get_bearing(lat, lon)
        bearing.attrs.update(dict(units='°'))

        # set geoinfo data set
        self._geo_infos = xr.Dataset({'dist': dist, 'bearing': bearing})

        # set wind_sectors
        # initialize WindSector() providing helper methods for wind sector calculations
        ws = WindSector(wind_sectors=self.wind_sectors, wind_sector_edge_dim_name=self.wind_sector_edge_dim_name)

        # expand dataset by new dataarray
        self._geo_infos['wind_sectors'] = xr.full_like(bearing, fill_value=np.nan) # xr.DataArray(coords=bearing.coords, dims=bearing.dims)
        for i, (k, v) in enumerate(ws.wind_sectore_edges.items()):
            self._geo_infos['wind_sectors'] = xr.where(
                ws.is_in_sector(k, self.geo_infos['bearing']), i, self._geo_infos['wind_sectors'])

    @property
    def geo_infos(self) -> xr.Dataset:
        return self._geo_infos

    def _apply_external_coordinates(self):
        """
        Apply external coordinates to data
        :return:
        :rtype:
        """
        ds_coords = xr.open_dataset(self.external_coords_file, chunks={'south_north': 36, 'west_east': 40})
        data = self._data
        for k, v in ds_coords.coords.variables.items():
            data = data.assign_coords({k: (remove_items(list(v.dims), 'Time'), v.values.squeeze())})
        self._data = data
        ds_coords.close()
        logging.debug('setup external coords')

    def _set_coords(self, coords):
        __set_coords = dict(lat=None, lon=None)
        if len(coords) != 2:
            raise SyntaxError(f"`coords' must have length=2 (lat, lon), but has length={len(coords)}")
        if isinstance(coords, tuple):
            self.__coords = dict(lat=coords[0], lon=coords[1])
            logging.debug(f"{self.__class__.__name__}.__coords is set to {self.__coords}")
        elif isinstance(coords, dict):
            if coords.keys() == __set_coords.keys():
                self.__coords = coords
            else:
                raise KeyError(f"dict `coords' must have the keys `lat', and `lon' but has: {coords.keys()}")
        else:
            raise TypeError(f"`coords' must be a tuple of floats or a dict, but is of type: {type(coords)}")

    def get_coordinates(self, as_arrays=False) -> Union[Tuple[np.ndarray, np.ndarray], dict]:
        """
        Retuns physical coordinates
        :param as_arrays: switch to return coords as array (True, lat/lon) or dictionary (False, default, lat/lon as keys)
        :return:
        :rtype:
        """
        if as_arrays:
            return np.array(self.__coords['lat']), np.array(self.__coords['lon'])
        else:
            return self.__coords

    def _set_nearest_icoords(self, dim=None):
        """
        Set nearest logical coordinates (icoords). Coordinates of gridbox center.
        :param dim:
        :type dim:
        :return:
        :rtype:
        """
        lat, lon = self.get_coordinates(as_arrays=True)
        with ProgressBar():
            logging.info("SingleGridColumnWrfChemDataLoader: compute nearest icoordinates")
            self._nearest_icoords = dask.compute(self.compute_nearest_icoordinates(lat, lon, dim))[0]

    def get_nearest_icoords(self, as_arrays=False):
        """
        Get nearest logical coordinates (icoords). Coordinates of gridbox center.
        :param as_arrays:
        :type as_arrays:
        :return:
        :rtype:
        """
        if as_arrays:
            return (self._nearest_icoords[self.logical_y_coord_name].values,
                    self._nearest_icoords[self.logical_x_coord_name].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_icoords.items()}

    def _set_nearest_coords(self):
        """
        Set nearest physical coordinates (coords). Coordinates of gridbox center.
        :return:
        :rtype:
        """
        icoords = self.get_nearest_icoords()
        ilat = convert2xrda(np.array(icoords[self.logical_y_coord_name]), use_1d_default=True)
        ilon = convert2xrda(np.array(icoords[self.logical_x_coord_name]), use_1d_default=True)
        lat = self._data[self.physical_y_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        lon = self._data[self.physical_x_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        self._nearest_coords = dict(lat=lat, lon=lon)

    def get_nearest_coords(self, as_arrays=False):
        """
        Set nearest physical coordinates (coords). Coordinates of gridbox center.

        :param as_arrays:
        :type as_arrays:
        :return:
        :rtype:
        """
        if as_arrays:
            return (self._nearest_coords['lat'].values,
                    self._nearest_coords['lon'].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_coords.items()}

    def apply_toarstats(self):
        dh = self.prepare_and_apply_toarstats(self.data, target_sampling=self.aggregation_sampling)
        return dh

    @staticmethod
    def _toarstats_wrapper(data, sampling, statistics, metadata, seasons=None, crops=None,
              data_capture=None):

        return toarstats(sampling, statistics, data, metadata, seasons, crops, data_capture)


    @TimeTrackingWrapper
    def prepare_and_apply_toarstats(self, data, target_sampling="daily"):
        meta = pd.DataFrame({self.physical_x_coord_name: self.get_coordinates()['lon'].tolist(),
                             self.physical_y_coord_name: self.get_coordinates()['lat'].tolist()},
                            index=self.station)
        if self.target_time_type is None:
            target_time_type_or_zone = self.get_local_time_zone_from_lat_lon(meta=meta)
        else:
            target_time_type_or_zone = self.target_time_type
        collector = []

        vars_for_toarstats = self.get_wrfnames_from_statistics_per_var()
        # for var in self.statistics_per_var.keys():
        for var, var_stats in vars_for_toarstats.items():
            collector.append(self._toarstats_aggregation(data, target_time_type_or_zone, meta, target_sampling, var, var_stats))
        sampling_data = xr.merge(collector).dropna(self.physical_t_coord_name)
        # sampling_data.attrs = data.attrs
        missing_squeezed_coords = data.coords._names - sampling_data.coords._names
        for coord in missing_squeezed_coords:
            sampling_data.coords[coord] = data.coords[coord]

        return sampling_data

    def _toarstats_aggregation(self, data, target_time_type_or_zone, meta, target_sampling, var, var_stats):
        with TimeTracking(name=f"{self.station}: apply toarstats `{var_stats}({var})`"):
            spatial_dims = list(remove_items(data[var].dims, self.physical_t_coord_name))
            try:
                df = data[var].to_dataframe()[[var]].reset_index(level=spatial_dims)
            except Exception as e:
                logging.info(f"can't reset_index in __toarstats_aggregation: {e}")
                df = data[var].to_dataframe()[[var]].reset_index(level=spatial_dims)

            df = df[[var] + spatial_dims]
            df = self.set_external_time_zone_and_convert_to_target_time(df, target_time_type_or_zone)
            df = df.groupby(spatial_dims)
            df = df.apply(self._toarstats_wrapper, sampling=target_sampling, statistics=var_stats,
                          metadata=(meta[self.physical_y_coord_name], meta[self.physical_x_coord_name]))
            df.columns = [var]
            df.index.set_names(df.index.names[:len(spatial_dims)] + [self.physical_t_coord_name], inplace=True)
            # df = df.to_xarray().to_array(self.target_dim)
            df = df.to_xarray()
            try:
                df = df.chunk({self.logical_x_coord_name: -1})
            except ValueError:
                df = df.chunk({self.logical_y_coord_name: -1})
            # collector.append(df)
            return df

    def set_external_time_zone_and_convert_to_target_time(self, data, target_time_type_or_zone):
        """

        :param data:
        :type data:
        :param target_time_type_or_zone:
        :type target_time_type_or_zone:
        :return:
        :rtype:
        """
        hdata = data
        # hdata = data.squeeze().to_pandas()
        if target_time_type_or_zone == "solar_time":
            lon = self.get_coordinates()['lon']
            toffset = dt.timedelta(hours=np.floor(lon/15.))
            hdata.index = hdata.index + toffset
        else:
            hdata.index = self.set_time_zone(hdata.index)
            hdata.index = hdata.index.tz_convert(target_time_type_or_zone)
            logging.debug(f"Set local time zone to: {target_time_type_or_zone}")
        return hdata

    def get_local_time_zone_from_lat_lon(self, lat=None, lon=None, meta=None):
        """
        Retuns name of time zone for given lat lon coordinates.

        Method also accepts a meta data pd.DataFrame where lat and lon are extracted with the loader's physical x and y
        coord names

        :param lat:
        :type lat:
        :param lon:
        :type lon:
        :param meta:
        :type meta: pd.DataFrame
        :return:
        :rtype: str
        """
        if (lat is None and lon is None) and (meta is not None):
            lat = meta[self.physical_y_coord_name]
            lon = meta[self.physical_x_coord_name]

        tz_where = tzwhere.tzwhere()
        local_time_zone = tz_where.tzNameAt(latitude=lat, longitude=lon)
        logging.debug(f"Detect local time zone '{local_time_zone}' based on lat={lat}, lon={lon}")
        return local_time_zone

    def set_time_zone(self, time_index):
        """
        Sets time zone information on a given index

        :param time_index:
        :type time_index:
        :return:
        :rtype:
        """

        dti = pd.to_datetime(time_index)
        dti = dti.tz_localize(self.time_zone)
        logging.debug(f"Set external time zone for {self.station} to: {self.time_zone}")
        return dti


class DataHandlerSingleGridColumn(DataHandlerSingleStation):
    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, external_coords_file=None,
                 var_logical_z_coord_selector=None,
                 targetvar_logical_z_coord_selector=None,
                 rechunk_values=None,
                 date_format_of_nc_file=None,
                 as_image_like_data_format=True,
                 time_zone=None,
                 target_time_type=None,
                 input_output_sampling4toarstats : tuple = None,
                 experiment_path: str = None,
                 vars_for_unit_conv: Dict = None,
                 **kwargs):
        self.external_coords_file = external_coords_file
        self.var_logical_z_coord_selector = self._return_z_coord_select_if_valid(var_logical_z_coord_selector,
                                                                                 as_input=True)
        self.targetvar_logical_z_coord_selector = self._return_z_coord_select_if_valid(targetvar_logical_z_coord_selector,
                                                                                       as_input=False)
        self._logical_z_coord_name = None
        self._joint_z_coord_selector = self._extract_largest_coord_extractor(self.var_logical_z_coord_selector,
                                                                             self.targetvar_logical_z_coord_selector)
        self.__loader = None
        self.rechunk_values = rechunk_values
        self.date_format_of_nc_file = date_format_of_nc_file
        self.as_image_like_data_format = as_image_like_data_format

        self.time_zone = time_zone
        self.target_time_type = target_time_type
        self.input_output_sampling4toarstats = input_output_sampling4toarstats
        self.experiment_path = experiment_path
        self.vars_for_unit_conv = vars_for_unit_conv
        super().__init__(*args, **kwargs)

    @staticmethod
    def _return_z_coord_select_if_valid(z_coord: int_or_list_of_int, as_input: bool) -> int_or_list_of_int:
        if isinstance(z_coord, int):
            return z_coord
        elif isinstance(z_coord, list):
            all_entries_int = all(isinstance(zi, int) for zi in z_coord)
            if all_entries_int and as_input:
                return z_coord
            else:
                raise NotImplementedError(
                    f"z_coord selector of type list not implemented for target vars. "
                    f"`all_entries_int'={all_entries_int}, `as_input'= {as_input}")
        else:
            raise TypeError(f"`z_coord' must be of type int or list[int], but is {type(z_coord)}")

    @staticmethod
    def coord_str2coords(str_coords: str, sep='__', dec_marker='_') -> Tuple[float_np_xr, float_np_xr]:
        """
        Converts string of station names including lat lon information to numpy array
        :param str_coords: Name of point of interest (pseudo station) e.g. 'coords__50_7536__7_0827'
        :param sep: separator between 'name', 'lat', and 'lon' within the coordinate string (e.g.default  '__')
        :param dec_marker: decimal marker for lat and lon values (e.g default '_'
        :type dec_marker:
        :return:
        :rtype:
        """
        if isinstance(str_coords, list) and len(str_coords) == 1:
            str_coords = str_coords[0]
        _, lat, lon = str_coords.split(sep=sep)
        lat = lat.replace(dec_marker, '.')
        lon = lon.replace(dec_marker, '.')

        return np.array(float(lat)), np.array(float(lon))

    def setup_data_path(self, data_path: str, sampling: str):
        # ToDo What is this good for?!?!?!
        return data_path

    @TimeTrackingWrapper
    def extract_data_from_loader(self, loader):
        data = loader.data.isel(loader.get_nearest_icoords()).squeeze(dim=[loader.logical_y_coord_name,
                                                                           loader.logical_x_coord_name])[self.variables]
        with ProgressBar():
            data = dask.compute(data)[0]
        return data

    @property
    def loader(self):
        return self.__loader

    @loader.setter
    def loader(self, station_path, **kwargs):
        try:
            station, path = station_path
        except ValueError as e:
            raise ValueError(f"Pass an iterable with two items; (station, path)")
        lat, lon = self.coord_str2coords(station)
        if hasattr(self, "wind_sectors"):
            helper_wind_sectors = self.wind_sectors
        else:
            helper_wind_sectors = None

        if hasattr(self, "wind_sector_edge_dim_name"):
            helper_wind_sector_edge_dim_name = self.wind_sector_edge_dim_name
        else:
            helper_wind_sector_edge_dim_name = None

        loader = SingleGridColumnWrfChemDataLoader((lat, lon),
                                                   data_path=path,
                                                   external_coords_file=self.external_coords_file,
                                                   time_dim_name=self.time_dim,
                                                   rechunk_values=self.rechunk_values,
                                                   variables=self.variables,
                                                   z_coord_selector=self._joint_z_coord_selector,
                                                   # start_time=self.start,
                                                   # end_time=self.end,
                                                   date_format_of_nc_file=self.date_format_of_nc_file,
                                                   wind_sectors=helper_wind_sectors,
                                                   wind_sector_edge_dim_name=helper_wind_sector_edge_dim_name,
                                                   statistics_per_var=self.statistics_per_var,
                                                   aggregation_sampling=self.input_output_sampling4toarstats[1],
                                                   time_zone=self.time_zone,
                                                   target_time_type=self.target_time_type,
                                                   station=self.station,
                                                   lazy_preprocessing=True,
                                                   vars_for_unit_conv=self.vars_for_unit_conv,
                                                   )
        self.__loader = loader

    def load_data(self, path, station, statistics_per_var, sampling, station_type=None, network=None,
                  store_data_locally=False, data_origin: Dict = None, start=None, end=None, meta_col_names=None):

        self.loader = (station, path)
        if meta_col_names is None:
            meta_col_names = ['station_name', 'station_lon', 'station_lat', 'station_alt']

        with self.loader as loader:
            if self._logical_z_coord_name is None:
                self._logical_z_coord_name = loader.logical_z_coord_name
            # # select defined variables at grid box or grid coloumn based on nearest icoords
            data = self.extract_data_from_loader(loader)
            if self._joint_z_coord_selector is not None:
                data = data.sel({self._logical_z_coord_name: self._joint_z_coord_selector})
            with ProgressBar(), TimeTracking(name=f"{self.station}: get data"):
                logging.info(f"start compute data for {self.station} in load_data")
                data = dask.compute(data)[0]

        # set time zone to attrs
        data.attrs["time_zone"] = self.time_zone

        # expand dimesion for iterdim
        data = data.expand_dims({self.iter_dim: station}).to_array(self.target_dim)
        # transpose dataarray: set first three fixed and keep remaining as is
        data = data.transpose(self.iter_dim, self.time_dim, self.target_dim, ...)


        # ToDo add metadata
        _meta = {self.loader.physical_x_coord_name: self.loader.get_coordinates()['lon'].tolist(),
                 self.loader.physical_y_coord_name: self.loader.get_coordinates()['lat'].tolist()
                 }
        _meta.update({k: v for k, v in zip(meta_col_names, (station[0],
                                                               _meta[self.loader.physical_x_coord_name],
                                                               _meta[self.loader.physical_y_coord_name], np.nan))
                      })
        meta = pd.DataFrame(_meta, index=station).T

        # if isinstance(self.input_output_sampling4toarstats, tuple) and len(self.input_output_sampling4toarstats) == 2:
        #     if self.var_logical_z_coord_selector != 0:
        #         raise NotImplementedError(
        #             f"Method `apply_toarstats` is not implemented for var_logical_z_coord_selector != 0: "
        #             f"Is {self.var_logical_z_coord_selector}")
            # data = self.prepare_and_apply_toarstats(data, meta, self.input_output_sampling4toarstats[1])
        data = self._slice_prep(data, start=start, end=end)

        return data, meta

    # @TimeTrackingWrapper
    # def prepare_and_apply_toarstats(self, data, meta, target_sampling="daily"):
    #     local_time_zone = self.get_local_time_zone_from_lat_lon(meta=meta)
    #     hdata = self.set_external_time_zone_and_convert_to_target_time(data, local_time_zone)
    #     hsampling_data = []
    #     for i, var in enumerate(hdata.columns):
    #         hdf = toarstats(target_sampling, self.statistics_per_var[var], hdata[var],
    #                         (meta[self.loader.physical_y_coord_name], meta[self.loader.physical_x_coord_name]))
    #         hsampling_data.append(xr.DataArray(hdf, coords=[hdf.index, [var]],
    #                                            dims=[self.loader.physical_t_coord_name, self.target_dim]))
    #     sampling_data = xr.concat(hsampling_data, dim=self.target_dim)
    #     sampling_data = sampling_data.broadcast_like(data, exclude=self.loader.physical_t_coord_name).dropna(
    #         self.loader.physical_t_coord_name)
    #     sampling_data.attrs = data.attrs
    #     missing_squeezed_coords = data.coords._names - sampling_data.coords._names
    #     for coord in missing_squeezed_coords:
    #         sampling_data.coords[coord] = data.coords[coord]
    #
    #     return self._force_dask_computation(sampling_data)
    #
    # def set_external_time_zone_and_convert_to_target_time(self, data, local_time_zone):
    #     """
    #
    #     :param data:
    #     :type data:
    #     :param local_time_zone:
    #     :type local_time_zone:
    #     :return:
    #     :rtype:
    #     """
    #     hdata = data.squeeze().to_pandas()
    #     hdata.index = self.set_time_zone(hdata.index)
    #     hdata.index = hdata.index.tz_convert(local_time_zone)
    #     logging.debug(f"Set local time zone for {self.station} to: {local_time_zone}")
    #     return hdata
    #
    # def get_local_time_zone_from_lat_lon(self, lat=None, lon=None, meta=None):
    #     """
    #     Retuns name of time zone for given lat lon coordinates.
    #
    #     Method also accepts a meta data pd.DataFrame where lat and lon are extracted with the loader's physical x and y
    #     coord names
    #
    #     :param lat:
    #     :type lat:
    #     :param lon:
    #     :type lon:
    #     :param meta:
    #     :type meta: pd.DataFrame
    #     :return:
    #     :rtype: str
    #     """
    #     if (lat is None and lon is None) and (meta is not None):
    #         lat = meta[self.loader.physical_y_coord_name]
    #         lon = meta[self.loader.physical_x_coord_name]
    #
    #
    #     tz_where = tzwhere.tzwhere()
    #     local_time_zone = tz_where.tzNameAt(latitude=lat, longitude=lon)
    #     logging.debug(f"Detect local time zone '{local_time_zone}' based on lat={lat}, lon={lon}")
    #     return local_time_zone
    #
    # def set_time_zone(self, time_index):
    #     """
    #     Sets time zone information on a given index
    #
    #     :param time_index:
    #     :type time_index:
    #     :return:
    #     :rtype:
    #     """
    #
    #     dti = pd.to_datetime(time_index)
    #     dti = dti.tz_localize(self.time_zone)
    #     logging.debug(f"Set external time zone for {self.station} to: {self.time_zone}")
    #     return dti

    @staticmethod
    def _extract_largest_coord_extractor(var_extarctor, target_extractor) -> Union[List, None]:
        if var_extarctor is not None and target_extractor is not None:
            res = list(set(to_list(var_extarctor) + to_list(target_extractor)))
        else:
            res = None
        return res

    def get_X(self, upsampling=False, as_numpy=False):
        x_data = self.get_transposed_history()
        if self.as_image_like_data_format is False:
            x_data = x_data.squeeze()
        if as_numpy is True:
            # return None
            raise NotImplementedError(f"keyword argument `as_numpy=True' not implemented.")
        elif as_numpy is False:
            return x_data

    # def get_Y(self, upsampling=False, as_numpy=False):
    #     raise NotImplementedError

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim, ...).copy()

    def get_transposed_label(self) -> xr.DataArray:
        """Return label.

        :return: label with dimensions datetime*, window*, Stations, variables.
        """
        return self.label.squeeze([self.iter_dim, self.target_dim]).transpose(self.time_dim, self.window_dim, ...).copy()

    @TimeTrackingWrapper
    def make_samples(self):
        self.make_history_window(self.target_dim, self.window_history_size, self.time_dim)
        if self.var_logical_z_coord_selector is not None:
            self.history = self.history.sel({self._logical_z_coord_name: self.var_logical_z_coord_selector})
        self.make_labels(self.target_dim, self.target_var, self.time_dim, self.window_lead_time)
        self.make_observation(self.target_dim, self.target_var, self.time_dim)
        if self.targetvar_logical_z_coord_selector is not None:
            self.label = self.label.sel({self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})
            self.observation = self.observation.sel(
                {self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})

        self.remove_nan(self.time_dim)

        self.history = self.modify_history()
        self.label = self.modify_label()
        self.observation = self.modify_observation()

        self.remove_nan(self.time_dim)

    def modify_history(self):
        """
        Place holder for more user spec. processing of history.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.history

    def modify_label(self):
        """
        Place holder for more user spec. processing of label.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.label

    def modify_observation(self, **kwargs):
        """
        Place holder for more user spec. processing of observation.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.observation

    @property
    def transformation_is_applied(self):
        """
        Returns True if transformation is applied.
        Checks if all entries in `self._transform' are not None/empty
        """
        return all(self._transformation)

    def interpolate(self, data, dim: str, method: str = 'linear', limit: int = None,
                    use_coordinate: Union[bool, str] = True, sampling="daily", **kwargs):
        """
        Interpolate values according to different methods.

        (Copy paste from dataarray.interpolate_na)

        :param dim:
                Specifies the dimension along which to interpolate.
        :param method:
                {'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                          'polynomial', 'barycentric', 'krog', 'pchip',
                          'spline', 'akima'}, optional
                    String indicating which method to use for interpolation:

                    - 'linear': linear interpolation (Default). Additional keyword
                      arguments are passed to ``numpy.interp``
                    - 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                      'polynomial': are passed to ``scipy.interpolate.interp1d``. If
                      method=='polynomial', the ``order`` keyword argument must also be
                      provided.
                    - 'barycentric', 'krog', 'pchip', 'spline', and `akima`: use their
                      respective``scipy.interpolate`` classes.
        :param limit:
                    default None
                    Maximum number of consecutive NaNs to fill. Must be greater than 0
                    or None for no limit.
        :param use_coordinate:
                default True
                    Specifies which index to use as the x values in the interpolation
                    formulated as `y = f(x)`. If False, values are treated as if
                    eqaully-spaced along `dim`. If True, the IndexVariable `dim` is
                    used. If use_coordinate is a string, it specifies the name of a
                    coordinate variariable to use as the index.
        :param kwargs:

        :return: xarray.DataArray
        """
        # data = self.create_full_time_dim(data, dim, sampling)
        return data.interpolate_na(dim=dim, method=method, limit=limit, use_coordinate=use_coordinate, **kwargs)



class DataHandlerWRF(DefaultDataHandler):
    """Data handler using DataHandlerSingleGridColumn."""
    data_handler = DataHandlerSingleGridColumn
    data_handler_transformation = DataHandlerSingleGridColumn
    _requirements = data_handler.requirements()


class DataHandlerSectorGrid(DataHandlerSingleGridColumn):
    # _requirements1 = remove_items(inspect.getfullargspec(DataHandlerWRF).args, ["self", "station"])
    # _requirements2 = remove_items(inspect.getfullargspec(DataHandlerSingleGridColumn).args, ["self", "station"])
    _requirements = DataHandlerWRF.requirements()

    def __init__(self, *args, radius=None, wind_sectors=None, wind_sector_edge_dim_name=None,  **kwargs):
        if radius is None:
            radius = 100  # km
        self.radius = radius
        if wind_sectors is None:
            wind_sectors = WindSector.DEFAULT_WIND_SECTORS
        self.wind_sectors = wind_sectors
        if wind_sector_edge_dim_name is None:
            wind_sector_edge_dim_name = 'edges'
        self.wind_sector_edge_dim_name = wind_sector_edge_dim_name
        self.windsector = WindSector(wind_sectors=self.wind_sectors,
                                     wind_sector_edge_dim_name=self.wind_sector_edge_dim_name)
        self._added_vars = []
        self.wind_dir_name = None
        self._wind_upstream_sector_by_name = None
        # self.wind_upstream_sector_by_name = None
        super().__init__(*args, **kwargs)

    def get_wind_upstream_sector_by_name(self):
        return self.wind_upstream_sector_by_name

    @property
    def wind_upstream_sector_by_name(self):
        return self._wind_upstream_sector_by_name

    @wind_upstream_sector_by_name.setter
    def wind_upstream_sector_by_name(self, wind_upstream_sector_by_name: xr.DataArray):
        self._wind_upstream_sector_by_name = wind_upstream_sector_by_name

    @TimeTrackingWrapper
    def _store_wind_upstream_sector_by_name(self):
        file_name = os.path.join(self.experiment_path,
                                 f"data/{self.station[0]}_{self.start}_{self.end}_upstream_wind_sector.nc")
        wind_upstream_sector_by_name = self.wind_upstream_sector_by_name
        dims_to_expand = list(wind_upstream_sector_by_name.coords._names - set(wind_upstream_sector_by_name.dims))
        wind_upstream_sector_by_name = wind_upstream_sector_by_name.expand_dims(dims_to_expand).squeeze(
            dim=list(set(dims_to_expand) - {self.iter_dim, self.target_dim}), drop=True)
        wind_upstream_sector_by_name = wind_upstream_sector_by_name.to_dataset(self.iter_dim)
        wind_upstream_sector_by_name.to_netcdf(file_name)
        logging.info(f"_store_wind_upstream_sector_by_name: {file_name}")

    @TimeTrackingWrapper
    def extract_data_from_loader(self, loader):
        wind_dir_name = self._get_wind_dir_var_name(loader)
        full_data = loader.data.isel(loader.get_nearest_icoords()).squeeze(dim=[loader.logical_y_coord_name,
                                                                                loader.logical_x_coord_name])
        data = full_data[self.variables]
        sec_data = self.windsector.get_sect_of_value(full_data[wind_dir_name])
        self.sec_data = sec_data
        if wind_dir_name not in data:
            data[wind_dir_name] = full_data[wind_dir_name]
            self._added_vars.append(to_list(wind_dir_name))
        with ProgressBar():
            data = dask.compute(data)[0]
        return data

    def _get_wind_dir_var_name(self, loader, wdir_name3d='wdirll', wdir_name2d='wdir10ll'):
        """
        Get variable name of wind direction. If no wind variable is given in self.variables, 10m wind is used as default.
        If a wind variable is used in self.variables, returns the variable name (e.g. wdirll if Ull or Vll is in self.variables,
        or wdir10ll if U10ll or V10ll is in self.variables. The same holds if wdirll/wspdll or wdir10ll/wspd10ll is given.

        :return:
        :rtype:
        """

        if (wdir_name3d in self.variables) and not (wdir_name2d in self.variables):
            var_name = wdir_name3d
        elif ((wdir_name2d in self.variables) and not (wdir_name3d in self.variables)) or (not (wdir_name2d in self.variables) and not (wdir_name3d in self.variables)):
            var_name = wdir_name2d
        else:
            raise ValueError(f"invalid wind var name passed to `_get_wind_dir_var_name'")

        assert var_name in itertools.chain(*itertools.chain(*loader.vars_to_rotate))
        if self.wind_dir_name is None:
            self.wind_dir_name = var_name
        return var_name

    def _get_left_and_right_wind_sector_edges(self, **kwargs):
        ws_edges = self.windsector.wind_sector_edges_data(**kwargs)
        ws_edges = ws_edges.expand_dims(dim={self.target_dim: to_list(self.wind_dir_name)})
        return ws_edges

    def get_transformation_opts(self, base=0):
        pos_of_trafo = self.get_transformation_base(base)
        trafo_opts = self._transformation[pos_of_trafo]
        return trafo_opts

    def apply_transformation_on_data(self, data):
        trafo_opts = self.get_transformation_opts()
        try:
            data = data.expand_dims(self._logical_z_coord_name)
        except ValueError as e:
            # ValueError: Dimension bottom_top already exists.
            pass
        trafo, _ = self.transform(data, dim=self.aggregation_dim, inverse=False,
                                  opts=trafo_opts, transformation_dim=self.target_dim)
        # except ValueError as e:
        #     pass # ToDo remove "bottom_top" from dims
        return trafo

    @TimeTrackingWrapper
    def modify_history(self):
        if self.transformation_is_applied:
            existing_sectors, sector_allocation, sector_history, sector_history_var_names = self.prepare_sector_allocation_and_history()
            with self.loader as loader, TimeTracking(name="loader in modify history"):
                # setup sector history
                grid_data = self.preselect_and_transform_neighbouring_data_based_on_radius(loader)
                for i, sect in enumerate(existing_sectors, start=1):
                    with TimeTracking(name=f"processing sector {sect} ({i}/{len(existing_sectors)}) for {self.station}"):
                        sector_history = self.get_wind_sectordata_in_history_format(grid_data, loader, sect,
                                                                                    sector_allocation, sector_history)

            sector_history = sector_history.assign_coords({self.target_dim: sector_history_var_names})
            combined_history = xr.concat([self.history, sector_history], dim=self.target_dim)

            return combined_history
        else:
            return self.history

    def prepare_sector_allocation_and_history(self):
        ws_edges = self.get_applied_transdormation_on_wind_sector_edges()
        wind_dir_of_interest = self.compute_wind_dir_of_interest()
        sector_allocation = self.windsector.get_sect_of_value(value=wind_dir_of_interest, external_edges=ws_edges)
        sector_allocation = sector_allocation.squeeze()
        existing_sectors = np.unique(sector_allocation.data)
        sector_history, sector_history_var_names = self.setup_history_like_xr_and_var_names()
        self.wind_upstream_sector_by_name = sector_allocation
        self._store_wind_upstream_sector_by_name()
        return existing_sectors, sector_allocation, sector_history, sector_history_var_names

    def setup_history_like_xr_and_var_names(self, var_name_suffix="Sect"):
        """
        Returns ones_like xarray from self.history and list of variable names which can be modified by passing a
        var_name_suffix

        :param var_name_suffix:
        :type var_name_suffix:
        :return:
        :rtype:
        """
        sector_history = xr.ones_like(self.history)
        sector_history_var_names = [f"{var}{var_name_suffix}" for var in sector_history.coords[self.target_dim].values]
        return sector_history, sector_history_var_names

    def get_wind_sectordata_in_history_format(self, grid_data, loader, sect, sector_allocation, sector_history,
                                              external_selector: dict = None):
        # select data in wind sector
        sec_data = self.get_section_data_from_circle(grid_data, loader, sect)
        sec_data = self.apply_aggregation_method_on_sector_data(sec_data, loader)
        sec_data_history = self._make_and_return_history_window(dim_name_of_shift=self.time_dim,
                                                                window=self.window_history_size,
                                                                data=sec_data.to_array(self.target_dim)
                                                                )
        sec_data_history = sec_data_history.squeeze().expand_dims(self.iter_dim)
        sec_data_history = sec_data_history.broadcast_like(self.history)
        sec_data_history = sec_data_history.transpose(*self.history.dims)
        if external_selector is None:
            time_index_of_sector = self.extract_time_index_of_sector(sect, sector_allocation, sector_history)
            selector = {self.time_dim: time_index_of_sector}
        else:
            selector = external_selector

        sector_history.loc[selector] = sec_data_history.sel(selector).values
        # sector_history.loc[{self.time_dim: time_index_of_sector}] = sec_data_history.sel(
        #     {self.time_dim: time_index_of_sector}).values

        return sector_history

    def extract_time_index_of_sector(self, sect, sector_allocation, sector_history):
        time_index_of_sector = sector_history[self.time_dim].where((sector_allocation.squeeze() == sect),
                                                                   drop=True).values
        return time_index_of_sector

    def get_section_data_from_circle(self, grid_data, loader, sect, compute=True):
        sec_data = grid_data.where(
            self.windsector.is_in_sector(sect, loader.geo_infos.bearing.drop('points').squeeze()))
        if compute:
            with ProgressBar():
                return dask.compute(sec_data)[0]
        else:
            return sec_data

    @TimeTrackingWrapper
    def preselect_and_transform_neighbouring_data_based_on_radius(self, loader, compute=True):
        """
        Select neighbouring grid boxes which have a maximal distance of pre-selected radius from full model field
        :param loader:
        :type loader:
        :return:
        :rtype:
        """
        # get data from loader
        grid_data = loader.data[self.variables]
        # select correct time steps and vertical layers
        grid_data = grid_data.sel({
            # self.time_dim: self.history.coords[self.time_dim].values,  # get data which is used in history
            self.time_dim: slice(self.start, self.end),  # get data which is used in history
            self._logical_z_coord_name: self.var_logical_z_coord_selector,  # get only used vertical layers
        })
        # select grid boxes which are closer than given radius
        grid_data = grid_data.where(
            loader.geo_infos.dist.sel({"points": 0}).drop("points") <= self.radius)
        # apply transformation on variables
        grid_data = self.apply_transformation_on_data(grid_data.to_array(self.target_dim)).to_dataset(self.target_dim)

        if compute:
            return dask.compute(grid_data)[0]
        else:
            return grid_data

    @staticmethod
    def apply_aggregation_method_on_sector_data(data, loader):
        data = data.mean(dim=(loader.logical_x_coord_name, loader.logical_y_coord_name))
        return data

    def compute_wind_dir_of_interest(self):
        wind_dir_of_intrest = self.history.sel({self.target_dim: self.wind_dir_name, self.window_dim: 0})
        return wind_dir_of_intrest

    def get_applied_transdormation_on_wind_sector_edges(self):
        ws_edges = self._get_left_and_right_wind_sector_edges(return_as='xr.da', dim=self.wind_sector_edge_dim_name)
        ws_edges = self.apply_transformation_on_data(ws_edges)
        return ws_edges


class DataHandlerMainSectWRF(DefaultDataHandler):
    """Data handler using DataHandlerSectorGrid."""
    data_handler = DataHandlerSectorGrid
    data_handler_transformation = DataHandlerSectorGrid
    _requirements = data_handler.requirements()


class DataHandler3SectorGrid(DataHandlerSectorGrid):
    _requirements = DataHandlerSectorGrid.requirements()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @TimeTrackingWrapper
    def modify_history(self):
        if self.transformation_is_applied:
            existing_sectors, sector_allocation, sector_history, sector_history_var_names = self.prepare_sector_allocation_and_history()
            sector_history_left, sector_history_var_names_left = self.setup_history_like_xr_and_var_names(var_name_suffix="SectLeft")
            sector_history_right, sector_history_var_names_right = self.setup_history_like_xr_and_var_names(var_name_suffix="SectRight")
            with self.loader as loader, TimeTracking(name="loader in modify history"):
                # setup sector history
                grid_data = self.preselect_and_transform_neighbouring_data_based_on_radius(loader)
                for i, sect in enumerate(existing_sectors, start=1):
                    sect_left, sect_right = self.get_left_and_right_sector_of_main(sect)
                    with TimeTracking(
                            name=f"processing sector {sect} ({i}/{len(existing_sectors)}) for {self.station}"):
                        sector_history = self.get_wind_sectordata_in_history_format(grid_data, loader, sect,
                                                                                    sector_allocation, sector_history)
                        time_index_of_sector = self.extract_time_index_of_sector(sect, sector_allocation, sector_history)

                        minor_selector = {self.time_dim: time_index_of_sector}
                        sector_history_left = self.get_wind_sectordata_in_history_format(grid_data, loader, sect_left,
                                                                                         sector_allocation,
                                                                                         sector_history_left,
                                                                                         external_selector=minor_selector)
                        sector_history_right = self.get_wind_sectordata_in_history_format(grid_data, loader, sect_right,
                                                                                          sector_allocation,
                                                                                          sector_history_right,
                                                                                          external_selector=minor_selector
                                                                                          )

            sector_history = sector_history.assign_coords({self.target_dim: sector_history_var_names})
            sector_history_left = sector_history.assign_coords({self.target_dim: sector_history_var_names_left})
            sector_history_right = sector_history.assign_coords({self.target_dim: sector_history_var_names_right})
            combined_history = xr.concat([self.history, sector_history,
                                          sector_history_left, sector_history_right],
                                         dim=self.target_dim)

            return combined_history
        else:
            return self.history

    def get_left_and_right_sector_of_main(self, main_sect):
        left_sect_idx = self.wind_sectors.index(main_sect) - 1
        # cyclic index: start at idx 0 if idx + 1 is larger then list of wind sectors
        right_sect_idx = (self.wind_sectors.index(main_sect) + 1) % len(self.wind_sectors)
        return self.wind_sectors[left_sect_idx], self.wind_sectors[right_sect_idx]


class DataHandlerMainMinorSectWRF(DefaultDataHandler):
    """Data handler using DataHandlerSectorGrid."""
    data_handler = DataHandler3SectorGrid
    data_handler_transformation = DataHandler3SectorGrid
    _requirements = data_handler.requirements()


if __name__ == '__main__':

    import cartopy.crs as ccrs
    from shapely.geometry import Point, Polygon
    import math
    from cartopy.geodesic import Geodesic
    # ToDo
    # test imports for circle plots
    import shapely


    def plot_map_proj(data, xlim=None, ylim=None, filename=None, point=None, radius=None, **kwargs):
        cmap = kwargs.pop('cmap', plt.cm.Reds)
        set_background = kwargs.pop('set_background', False)
        edgecolors = kwargs.pop('edgecolors', 'face')
        circle = kwargs.pop('circle', False)

        crs_latlon = ccrs.PlateCarree()
        if ylim is None:
            ylim = [-90, 90]
        if xlim is None:
            xlim = [-180, 180]
        if filename is None:
            filename = 'test_fig.pdf'
        if point is not None:
            lat, lon = point
        else:
            lat, lon = None, None
        # plt.figure(figsize=(14, 6))
        plt.figure()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        pm = data.squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                       cmap=cmap, edgecolors=edgecolors,
                                       **kwargs)
        ax.coastlines(resolution='50m')


        ##### Idea to draw circle from https://stackoverflow.com/a/58735566
        #### own try:
        if circle and ((lon is not None) and (lat is not None)):
            circle_points = Geodesic().circle(lon=lon, lat=lat,
                                              radius=radius*1000., # radius_in_meters,
                                              )
            geom = shapely.geometry.Polygon(circle_points)
            ax.add_geometries((geom,), crs=ccrs.PlateCarree(), facecolor='none', edgecolor='blue', linewidth=1)
        #
        #####

        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        if (lon is not None) and (lat is not None):
            plt.plot(lon, lat, '+k', markersize=7, transform=crs_latlon)
        if set_background:
            ax.stock_img()
        # plt.tight_layout()
        plt.savefig(filename, bbox_inches='tight')
        plt.close('all')

    def dummy_plot(data, skipna=None, xlim=None, ylim=None):
        plt.figure()
        crs_latlon = ccrs.PlateCarree()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        data.sum(dim='points',skipna=skipna).squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                     cmap=plt.cm.Reds, )
        ax.coastlines(resolution='50m')
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        plt.tight_layout()
        plt.savefig("Example_Multiple_points_test.pdf")
        print("plot multiple points")

    def sector(center, start_angle, end_angle, radius, steps=200):
        def polar_point(origin_point, angle, distance):
            return [origin_point.x + math.sin(math.radians(angle)) * distance,
                    origin_point.y + math.cos(math.radians(angle)) * distance]

        if start_angle > end_angle:
            start_angle = start_angle - 360
        else:
            pass
        step_angle_width = (end_angle - start_angle) / steps
        sector_width = (end_angle - start_angle)
        segment_vertices = []

        segment_vertices.append(polar_point(center, 0, 0))
        segment_vertices.append(polar_point(center, start_angle, radius))

        for z in range(1, steps):
            segment_vertices.append((polar_point(center, start_angle + z * step_angle_width, radius)))
        segment_vertices.append(polar_point(center, start_angle + sector_width, radius))
        segment_vertices.append(polar_point(center, 0, 0))
        return Polygon(segment_vertices)

    # lat_np = np.array([50.73333, 47.73333])
    # lon_np = np.array([7.1, 8.6])
    # 48_8479__10_0963
    # 51_4065__6_9660
    lat_np = np.array([51.4065])
    lon_np = np.array([6.9660])

    lat_xr = xr.DataArray(lat_np, dims=["points"], coords={'points': range(len(lat_np))})
    lon_xr = xr.DataArray(lon_np, dims=["points"], coords={'points': range(len(lon_np))})

    use_first_dummy_dataset = True
    if use_first_dummy_dataset:

        wrf_new = SingleGridColumnWrfChemDataLoader((lat_xr, lon_xr),
                                                    # data_path='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/2009/',
                                                    data_path="/media/felix/INTENSO/WRF_CHEM/JFM_2009/",
                                                    common_file_starter='wrfout_d01_',
                                                    time_dim_name='XTIME',
                                                    logical_x_coord_name='west_east',
                                                    logical_y_coord_name='south_north',
                                                    logical_z_coord_name='bottom_top',
                                                    rechunk_values={'XTIME': 24, 'bottom_top': 10},
                                                    # external_coords_file='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/coords.nc',
                                                    external_coords_file="/media/felix/INTENSO/WRF_CHEM/monthly/coords.nc",
                                                    wind_sectors=['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'],
                                                    )
        windsect = WindSector(wind_sectors=['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'])
        radius_from_point = 100.  # in km

        x_de_extract = [2, 15]
        y_de_extract = [47, 59]

        x_eu_extract = [-4, 20]
        y_eu_extract = [40, 60]

        with wrf_new as loader:
            o3 = loader.data.o3

        plot_map_proj(data=o3.isel({'XTIME': 0, 'bottom_top': 0}), xlim=x_de_extract, ylim=y_de_extract,
                      filename="ozone_map_eur.pdf", cbar_kwargs={'orientation': 'horizontal', 'pad': 0.01},
                      circle=False)
        icoords = dask.compute(wrf_new.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_np = wrf_new.get_distances(lat_np, lon_np)
        dist_xr = wrf_new.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))

        bearing_np = wrf_new.get_bearing(lat_np, lon_np)
        bearing_xr = wrf_new.get_bearing(lat_xr, lon_xr)
        bearing_xr.attrs.update(dict(units='°'))

        geo_info = xr.Dataset({'dist': dist_xr, 'bearing': bearing_xr})


        ###test for rotation
        vecrot = VectorRotateLambertConformal2latlon(xlat=wrf_new.data.XLAT.data, xlong=wrf_new.data.XLONG.data)
        ugrd = vecrot.interpolate_to_grid_center(wrf_new.data.U, 'west_east_stag')
        vgrd = vecrot.interpolate_to_grid_center(wrf_new.data.V, 'south_north_stag')
        ull_grd, vll_grd = vecrot.ugrd_vgrd2ull_vll(ugrd, vgrd)
        ull_stg, vll_stg = vecrot.ustg_vstg2ull_vll(wrf_new.data.U, wrf_new.data.V,
                                                    ustg_dim='west_east_stag', vstg_dim='south_north_stag')

        wspd_grd, wdir_grd = vecrot.ugrd_vgrd2wspd_wdir(ugrd, vgrd)
        wspd_stg, wdir_stg = vecrot.ustg_vstg2wspd_wdir(wrf_new.data.U, wrf_new.data.V,
                                                        ustg_dim='west_east_stag', vstg_dim='south_north_stag')
        ### end test rotation

        point = [wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]]
        for i, (data, xlim, ylim, point, kwargs) in enumerate(((wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}),
                                                                x_eu_extract, y_eu_extract, point,
                                                         {'circle': True,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}),
                                                         x_eu_extract, y_eu_extract, point,
                                                         {'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.dist, x_eu_extract, y_eu_extract, point,
                                                         {'cmap': plt.cm.get_cmap('Greys'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.bearing, x_eu_extract, y_eu_extract, point,
                                                         {'cmap': plt.cm.get_cmap('twilight'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.wind_sectors, x_eu_extract, y_eu_extract, point,
                                                         {
                                                            'cmap': plt.cm.get_cmap('tab20', len(wrf_new.wind_sectors)),
                                                            'circle': False,
                                                            'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}
                                                            #'set_background': True,
                                                            # 'levels': 16,
                                                           }),
                                                        (wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         x_eu_extract, y_eu_extract, point,
                                                         {'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'circle': True,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01}}),
                                                        (geo_info.dist.where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('Greys'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01,
                                                                          'shrink': .6}}),
                                                        (wrf_new.geo_infos.bearing.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('twilight'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .65}}),
                                                        (wrf_new.geo_infos.wind_sectors.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract,y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('tab20', len(wrf_new.wind_sectors)),
                                                                    'circle': False,
                                                                    'cbar_kwargs': {'orientation': 'horizontal',
                                                                                    'shrink': .6, 'pad': 0.01,
                                                                                    }
                                                                    #'set_background': True,
                                                                    }),
                                                        (wrf_new.data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         x_de_extract, y_de_extract, None,
                                                         {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal',
                                                                                                     'fraction':.1, 'pad': 0.01,
                                                                                                         }}),
                                                        )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          # point=[lat_np, lon_np],
                          # point=[wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]],
                          point=point,
                          filename=f'Example_sector{i}.pdf',
                          radius=radius_from_point,
                          # cbar_kwargs={'orientation': 'horizontal'},
                          **kwargs,
                          )
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          # point=[lat_np, lon_np],
                          # point=[wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]],
                          point=point,
                          filename=f'Example_sector{i}.png',
                          radius=radius_from_point,
                          # cbar_kwargs={'orientation': 'horizontal'},
                          **kwargs,
                          )

        for i, (data, xlim, ylim) in enumerate(
                ((wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}), [-42, 66], [23, 80]),
                 (geo_info.dist, [-42, 66], [23, 80]),
                 (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}).where(
                     dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), x_de_extract, y_de_extract),
                 (geo_info.dist.where(dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), x_de_extract, y_de_extract),
                 )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np],
                          filename=f'ExampleO3_dist{i}.pdf',
                          radius=radius_from_point)

    ######################### # Larger 4D data
    use_second_dummy_dataset = False
    if use_second_dummy_dataset:
        wrf_dh_4d = BaseWrfChemDataLoader(data_path='/home/felix/Data/WRF-Chem/upload_aura/2009/',
                                          common_file_starter='wrfout_d01_2009', time_dim_name='Time')
        wrf_dh_4d.open_data()
        wrf_dh_4d.rechunk_data({"XTIME": 1, "bottom_top": 34, "south_north": 36, "west_east": 40})
        lat_np = np.array([50.73333])
        lon_np = np.array([7.1])
        wrf_dh_4d._data = wrf_dh_4d._data.assign_coords(wrf_dh_4d._data.coords)
        icoords = dask.compute(wrf_dh_4d.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_xr = wrf_dh_4d.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))
        geo_info = xr.Dataset({'dist': dist_xr})

        for i, (data, xlim, ylim) in enumerate(((wrf_dh_4d._data.T2, [-42, 66], [23, 80]),
                                                (geo_info.dist, [-42, 66], [23, 80]),
                                                (wrf_dh_4d._data.T2.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), x_de_extract,
                                                 y_de_extract),
                                                (geo_info.dist.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), x_de_extract,
                                                 y_de_extract),
                                                )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np], filename=f'test_dist{i}.pdf')


    print()
