import gc
import warnings
from typing import Union, Callable, Tuple, Dict, Any
import logging

import datetime
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy import signal
import xarray as xr
import dask.array as da

from mlair.helpers import to_list, TimeTrackingWrapper, TimeTracking


class FIRFilter:
    from mlair.plotting.data_insight_plotting import PlotFirFilter

    def __init__(self, data, fs, order, cutoff, window, var_dim, time_dim, display_name=None, minimum_length=None,
                 offset=0, plot_path=None, plot_dates=None):
        self._filtered = []
        self._h = []
        self.data = data
        self.fs = fs
        self.order = order
        self.cutoff = cutoff
        self.window = window
        self.var_dim = var_dim
        self.time_dim = time_dim
        self.display_name = display_name
        self.minimum_length = minimum_length
        self.offset = offset
        self.plot_path = plot_path
        self.plot_dates = plot_dates
        self.run()

    def run(self):
        logging.info(f"{self.display_name}: start {self.__class__.__name__}")
        filtered = []
        h = []
        input_data = self.data.__deepcopy__()

        # collect some data for visualization
        if self.plot_dates is None:
            plot_pos = np.array([0.25, 1.5, 2.75, 4]) * 365 * self.fs
            self.plot_dates = [input_data.isel({self.time_dim: int(pos)}).coords[self.time_dim].values
                               for pos in plot_pos if pos < len(input_data.coords[self.time_dim])]
        plot_data = []

        for i in range(len(self.order)):
            # apply filter
            fi, hi = self.fir_filter(input_data, self.fs, self.cutoff[i], self.order[i], time_dim=self.time_dim,
                                     var_dim=self.var_dim, window=self.window, display_name=self.display_name)
            filtered.append(fi)
            h.append(hi)

            # visualization
            plot_data.append(self.create_visualization(fi, input_data, self.plot_dates, self.time_dim, self.fs, hi,
                                                       self.minimum_length, self.order, i, self.offset, self.var_dim))
            # calculate residuum
            input_data = input_data - fi

        # add last residuum to filtered
        filtered.append(input_data)

        self._filtered = filtered
        self._h = h

        # visualize
        if self.plot_path is not None:
            try:
                self.PlotFirFilter(self.plot_path, plot_data, self.display_name)  # not working when t0 != 0
            except Exception as e:
                logging.info(f"Could not plot climate fir filter due to following reason:\n{e}")

    def create_visualization(self, filtered, filter_input_data, plot_dates, time_dim, sampling,
                              h, minimum_length, order, i, offset, var_dim):  # pragma: no cover
        plot_data = []
        minimum_length = minimum_length or 0
        for viz_date in set(plot_dates).intersection(filtered.coords[time_dim].values):
            try:
                if i < len(order) - 1:
                    minimum_length += order[i+1]

                td_type = {1: "D", 24: "h"}.get(sampling)
                length = len(h)
                extend_length_history = minimum_length + int((length + 1) / 2)
                extend_length_future = int((length + 1) / 2) + 1
                t_minus = viz_date + np.timedelta64(int(-extend_length_history), td_type)
                t_plus = viz_date + np.timedelta64(int(extend_length_future + offset), td_type)
                time_slice = slice(t_minus, t_plus - np.timedelta64(1, td_type))
                plot_data.append({"t0": viz_date, "filter_input": filter_input_data.sel({time_dim: time_slice}),
                                  "filtered": filtered.sel({time_dim: time_slice}), "h": h, "time_dim": time_dim,
                                  "var_dim": var_dim})
            except:
                pass
        return plot_data

    @property
    def filter_coefficients(self):
        return self._h

    @property
    def filtered_data(self):
        return self._filtered

    @TimeTrackingWrapper
    def fir_filter(self, data, fs, cutoff_high, order, sampling="1d", time_dim="datetime", var_dim="variables", window: Union[str, Tuple] = "hamming",
                   minimum_length=None, new_dim="window", plot_dates=None, display_name=None):

        # calculate FIR filter coefficients
        h = self._calculate_filter_coefficients(window, order, cutoff_high, fs)

        coll = []
        for var in data.coords[var_dim]:
            d = data.sel({var_dim: var})
            filt = xr.apply_ufunc(fir_filter_convolve, d,
                                  input_core_dims=[[time_dim]], output_core_dims=[[time_dim]],
                                  vectorize=True, kwargs={"h": h}, output_dtypes=[d.dtype])
            coll.append(filt)
        filtered = xr.concat(coll, var_dim)

        # create result array with same shape like input data, gaps are filled by nans
        filtered = self._create_full_filter_result_array(data, filtered, time_dim, display_name)
        return filtered, h

    @staticmethod
    def _calculate_filter_coefficients(window: Union[str, tuple], order: Union[int, tuple], cutoff_high: float,
                                       fs: float) -> np.array:
        """
        Calculate filter coefficients for moving window using scipy's signal package for common filter types and local
        method firwin_kzf for Kolmogorov Zurbenko filter (kzf). The filter is a low-pass filter.

        :param window: name of the window type which is either a string with the window's name or a tuple containing the
            name but also some parameters (e.g. `("kaiser", 5)`)
        :param order: order of the filter to create as int or parameters m and k of kzf
        :param cutoff_high: cutoff frequency to use for low-pass filter in frequency of fs
        :param fs: sampling frequency of time series
        """
        if window == "kzf":
            h = firwin_kzf(*order)
        else:
            h = signal.firwin(order, cutoff_high, pass_zero="lowpass", fs=fs, window=window)
        return h

    @staticmethod
    def _create_full_filter_result_array(template_array: xr.DataArray, result_array: xr.DataArray, new_dim: str,
                                         display_name: str = None) -> xr.DataArray:
        """
        Create result filter array with same shape line given template data (should be the original input data before
        filtering the data). All gaps are filled by nans.

        :param template_array: this array is used as template for shape and ordering of dims
        :param result_array: array with data that are filled into template
        :param new_dim: new dimension which is shifted/appended to/at the end (if present or not)
        :param display_name: string that is attached to logging (default None)
        """
        logging.debug(f"{display_name}: create res_full")
        new_coords = {**{k: template_array.coords[k].values for k in template_array.coords if k != new_dim},
                      new_dim: result_array.coords[new_dim]}
        dims = [*template_array.dims, new_dim] if new_dim not in template_array.dims else template_array.dims
        result_array = result_array.transpose(*dims)
        return result_array.broadcast_like(xr.DataArray(dims=dims, coords=new_coords))


class ClimateFIRFilter(FIRFilter):
    from mlair.plotting.data_insight_plotting import PlotClimateFirFilter

    def __init__(self, data, fs, order, cutoff, window, time_dim, var_dim, apriori=None, apriori_type=None,
                 apriori_diurnal=False, sel_opts=None, plot_path=None,
                 minimum_length=None, new_dim=None, display_name=None, extend_length_opts: int = 0,
                 offset: Union[dict, int] = 0, plot_dates=None):
        """
        :param data: data to filter
        :param fs: sampling frequency in 1/days -> 1d: fs=1 -> 1H: fs=24
        :param order: a tuple with the order of the filter in same ordering like cutoff
        :param cutoff: a tuple with the cutoff frequencies (all are applied as low pass)
        :param window: window type of the filter (e.g. hamming)
        :param time_dim: name of time dimension to apply filter along
        :param var_dim: name of variables dimension
        :param apriori: apriori information to use for the first low pass. If None, climatology is calculated on the
            provided data.
        :param apriori_type: type of apriori information to use. Climatology will be used always for first low pass. For
            the residuum either the value zero is used (apriori_type is None or "zeros") or a climatology on the
            residua is used ("residuum_stats").
        :param apriori_diurnal: Use diurnal cycle as additional apriori information (only applicable for hourly
            resoluted data). The mean anomaly of each hour is added to the apriori_type information.
        :param sel_opts: specify some parameters to select a subset of data before calculating the apriori information.
            Use this parameter for example, if apriori shall only calculated on a shorter time period than available in
            given data.
        :param extend_length_opts: shift information switch between historical data and apriori estimation by the given
            values (default None). Must either be a dictionary with keys available in var_dim or a single value that is
            applied to all data. This parameter has only influence on which information is available at t0 for the
            filter calculcation but has no influence on the shape of the returned filtered data.
        :param offset: This parameter indicates the number of time steps with ti>t0 to return of the filtered data. In
            case the offset parameter is larger than the extend_lenght_opts parameter, this leads to the case that not
            only observational data but also climatological estimations are returned.  Must either be a dictionary with
            keys available in var_dim or a single value that is applied to all data. Default is 0.
        """
        self._apriori = apriori
        self.apriori_type = apriori_type
        self.apriori_diurnal = apriori_diurnal
        self._apriori_list = []
        self.sel_opts = sel_opts
        self.new_dim = new_dim
        self.plot_data = []
        self.extend_length_opts = extend_length_opts
        super().__init__(data, fs, order, cutoff, window, var_dim, time_dim, display_name=display_name,
                         minimum_length=minimum_length, plot_path=plot_path, offset=offset, plot_dates=plot_dates)

    def run(self):
        filtered = []
        h = []
        if self.sel_opts is not None:
            self.sel_opts = self.sel_opts if isinstance(self.sel_opts, dict) else {self.time_dim: self.sel_opts}
        sampling = {1: "1d", 24: "1H"}.get(int(self.fs))
        logging.debug(f"{self.display_name}: create diurnal_anomalies")
        if self.apriori_diurnal is True and sampling == "1H":
            diurnal_anomalies = self.create_seasonal_hourly_mean(self.data, self.time_dim, sel_opts=self.sel_opts,
                                                                 sampling=sampling, as_anomaly=True)
        else:
            diurnal_anomalies = 0
        logging.debug(f"{self.display_name}: create monthly apriori")
        if self._apriori is None:
            self._apriori = self.create_monthly_mean(self.data, self.time_dim, sel_opts=self.sel_opts,
                                                     sampling=sampling) + diurnal_anomalies
            logging.debug(f"{self.display_name}: apriori shape = {self._apriori.shape}")
        apriori_list = to_list(self._apriori)
        input_data = self.data.__deepcopy__()

        # for visualization
        plot_dates = self.plot_dates

        # create tmp dimension to apply filter, search for unused name
        new_dim = self._create_tmp_dimension(input_data) if self.new_dim is None else self.new_dim

        for i in range(len(self.order)):
            logging.info(f"{self.display_name}: start filter for order {self.order[i]}")
            # calculate climatological filter
            next_order = self._next_order(self.order, 0, i, self.window)
            fi, input_data, hi, apriori, plot_data = self.clim_filter(input_data, self.fs, self.cutoff[i], self.order[i],
                                                                      apriori=apriori_list[i], sel_opts=self.sel_opts,
                                                                      sampling=sampling, time_dim=self.time_dim,
                                                                      window=self.window, var_dim=self.var_dim,
                                                                      minimum_length=self.minimum_length, new_dim=new_dim,
                                                                      plot_dates=plot_dates, display_name=self.display_name,
                                                                      extend_opts=self.extend_length_opts,
                                                                      offset=self.offset, next_order=next_order)

            logging.info(f"{self.display_name}: finished clim_filter calculation")
            if self.minimum_length is None:
                filtered.append(fi.sel({new_dim: slice(None, self.offset)}))
            else:
                filtered.append(fi.sel({new_dim: slice(self.offset - self.minimum_length, self.offset)}))
            h.append(hi)
            gc.collect()
            self.plot_data.append(plot_data)
            plot_dates = {e["t0"] for e in plot_data}

            # calculate residuum
            logging.info(f"{self.display_name}: calculate residuum")
            coord_range = range(fi.coords[new_dim].values.min(), fi.coords[new_dim].values.max() + 1)
            if new_dim in input_data.coords:
                input_data = input_data.sel({new_dim: coord_range}) - fi
            else:
                input_data = self._shift_data(input_data, coord_range, self.time_dim, new_dim) - fi

            # create new apriori information for next iteration if no further apriori is provided
            if len(apriori_list) < len(self.order):
                logging.info(f"{self.display_name}: create diurnal_anomalies")
                if self.apriori_diurnal is True and sampling == "1H":
                    diurnal_anomalies = self.create_seasonal_hourly_mean(input_data.sel({new_dim: 0}, drop=True),
                                                                         self.time_dim, sel_opts=self.sel_opts,
                                                                         sampling=sampling, as_anomaly=True)
                else:
                    diurnal_anomalies = 0
                logging.info(f"{self.display_name}: create monthly apriori")
                if self.apriori_type is None or self.apriori_type == "zeros":  # zero version
                    apriori_list.append(xr.zeros_like(apriori_list[i]) + diurnal_anomalies)
                elif self.apriori_type == "residuum_stats":  # calculate monthly statistic on residuum
                    apriori_list.append(
                        -self.create_monthly_mean(input_data.sel({new_dim: 0}, drop=True), self.time_dim,
                                                  sel_opts=self.sel_opts, sampling=sampling) + diurnal_anomalies)
                else:
                    raise ValueError(f"Cannot handle unkown apriori type: {self.apriori_type}. Please choose from None,"
                                     f" `zeros` or `residuum_stats`.")

        # add last residuum to filtered
        if self.minimum_length is None:
            filtered.append(input_data.sel({new_dim: slice(None, self.offset)}))
        else:
            filtered.append(input_data.sel({new_dim: slice(self.offset - self.minimum_length, self.offset)}))

        self._filtered = filtered
        self._h = h
        self._apriori_list = apriori_list

        # visualize
        if self.plot_path is not None:
            try:
                self.PlotClimateFirFilter(self.plot_path, self.plot_data, sampling, self.display_name)
            except Exception as e:
                logging.info(f"Could not plot climate fir filter due to following reason:\n{e}")

    @staticmethod
    def _next_order(order: list, minimum_length: Union[int, None], pos: int, window: Union[str, tuple]) -> int:
        next_order = 0
        if pos + 1 < len(order):
            next_order = order[pos + 1]
            if window == "kzf" and isinstance(next_order, tuple):
                next_order = filter_width_kzf(*next_order)
        if minimum_length is not None:
            next_order = next_order + minimum_length
        return next_order

    @staticmethod
    def create_monthly_unity_array(data: xr.DataArray, time_dim: str, extend_range: int = 366) -> xr.DataArray:
        """
        Create a xarray data array filled with ones with monthly resolution (set on 16th of month). Data is extended by
        extend_range days in future and past along time_dim.

        :param data: data to create monthly unity array from, must contain dimension time_dim
        :param time_dim: name of temporal dimension
        :param extend_range: number of days to extend data (default 366)
        :returns: xarray in monthly resolution (centered at 16th day of month) with all values equal to 1
        """
        coords = data.coords

        # extend time_dim by given extend_range days
        start = coords[time_dim][0].values.astype("datetime64[D]") - np.timedelta64(extend_range, "D")
        end = coords[time_dim][-1].values.astype("datetime64[D]") + np.timedelta64(extend_range, "D")
        new_time_axis = np.arange(start, end).astype("datetime64[ns]")

        # construct data array with updated coords
        new_coords = {k: data.coords[k].values if k != time_dim else new_time_axis for k in coords}
        new_array = xr.DataArray(1, coords=new_coords, dims=new_coords.keys()).transpose(*data.dims)

        # loffset is required because resampling uses last day in month as resampling timestamp
        return new_array.resample({time_dim: "1m"}, loffset=datetime.timedelta(days=-15)).max()

    def create_monthly_mean(self, data: xr.DataArray, time_dim: str, sel_opts: dict = None, sampling: str = "1d") \
            -> xr.DataArray:
        """
        Calculate monthly means (12 values) and return a data array with same resolution as given data containing these
        monthly mean values. Sampling points are the 16th of each month (this value is equal to the true monthly mean)
        and all other values between two points are interpolated linearly. It is possible to apply some pre-selection
        to use only a subset of given data using the sel_opts parameter. Only data from this subset are used to
        calculate the monthly statistic.

        :param data: data to apply statistical calculation on
        :param time_dim: name of temporal axis
        :param sel_opts: selection options as dict to select a subset of data (default None). A given sel_opts with
            `sel_opts={<time_dim>: "2006"}` forces the method e.g. to derive the monthly means only from data of the
            year 2006.
        :param sampling: sampling of the returned data (default 1d)
        :returns: array in desired resolution containing interpolated monthly values. Months with no valid data are
             returned as np.nan which also effects data in the neighbouring months (before / after sampling points which
             are the 16th of each month).
        """

        # create unity xarray in monthly resolution with sampling point in mid of each month
        monthly = self.create_monthly_unity_array(data, time_dim) * np.nan

        # apply selection if given (only use subset for monthly means)
        if sel_opts is not None:
            data = data.sel(**sel_opts)

        # create monthly mean and replace entries in unity array
        monthly_mean = data.groupby(f"{time_dim}.month").mean()
        for month in monthly_mean.month.values:
            monthly = xr.where((monthly[f"{time_dim}.month"] == month),
                               monthly_mean.sel(month=month, drop=True),
                               monthly)
        # transform monthly information into original sampling rate
        return monthly.resample({time_dim: sampling}).interpolate()

    @staticmethod
    def _compute_hourly_mean_per_month(data: xr.DataArray, time_dim: str, as_anomaly: bool) -> Dict[int, xr.DataArray]:
        """
        Calculate for each hour in each month a separate mean value (12 x 24 values in total). Average is either the
        anomaly of a monthly mean state or the raw mean value.

        :param data: data to calculate averages on
        :param time_dim: name of temporal dimension
        :param as_anomaly: indicates whether to calculate means as anomaly of a monthly mean or as raw mean values.
        :returns: dictionary containing 12 months each with a 24-valued array (1 entry for each hour)
        """
        seasonal_hourly_means = {}
        for month in data.groupby(f"{time_dim}.month").groups.keys():
            single_month_data = data.sel({time_dim: (data[f"{time_dim}.month"] == month)})
            hourly_mean = single_month_data.groupby(f"{time_dim}.hour").mean()
            if as_anomaly is True:
                hourly_mean = hourly_mean - hourly_mean.mean("hour")
            seasonal_hourly_means[month] = hourly_mean
        return seasonal_hourly_means

    @staticmethod
    def _create_seasonal_cycle_of_single_hour_mean(result_arr: xr.DataArray, means: Dict[int, xr.DataArray], hour: int,
                                                   time_dim: str, sampling: str) -> xr.DataArray:
        """
        Use monthly means of a given hour to create an array with interpolated values at the indicated hour for each day
        of the full time span indicated by given result_arr.

        :param result_arr: template array indicating the full time range and additional dimensions to keep
        :param means: dictionary containing 24 hourly averages for each month (12 x 24 values in total)
        :param hour: integer of hour of interest
        :param time_dim: name of temporal dimension
        :param sampling: sampling rate to interpolate
        :returns: array with interpolated averages in sampling resolution containing only values for hour of interest
        """
        h_coll = xr.ones_like(result_arr) * np.nan
        for month in means.keys():
            hourly_mean_single_month = means[month].sel(hour=hour, drop=True)
            h_coll = xr.where((h_coll[f"{time_dim}.month"] == month), hourly_mean_single_month, h_coll)
        h_coll = h_coll.resample({time_dim: sampling}).interpolate()
        h_coll = h_coll.sel({time_dim: (h_coll[f"{time_dim}.hour"] == hour)})
        return h_coll

    def create_seasonal_hourly_mean(self, data: xr.DataArray, time_dim: str, sel_opts: Dict[str, Any] = None,
                                    sampling: str = "1H", as_anomaly: bool = True) -> xr.DataArray:
        """
        Compute climatological statistics on hourly base either as raw data or anomalies. For each month, an overall
        mean value (only used if requiring anomalies) and the mean of each hour are calculated. The climatological
        diurnal cycle is positioned on the 16th of each month and interpolated in between by using a distinct
        interpolation for each hour of day. The returned array therefore contains data with a yearly cycle (if anomaly
        is not calculated) or data without a yearly cycle (if using anomalies). In both cases, the data have an
        amplitude that varies over the year.

        :param data: data to apply this method to
        :param time_dim: name of temporal axis
        :param sel_opts: specific selection options that are applied before calculation of climatological statistics
            (default None)
        :param sampling: temporal resolution of data (default "1H")
        :param as_anomaly: specify whether to use anomalies or raw data including a seasonal cycle of the mean value
            (default: True)
        :returns: climatological statistics for given data interpolated with given sampling rate
        """

        """Calculate hourly statistics. Either the absolute value or the anomaly (as_anomaly=True)."""
        # can only be used for hourly sampling rate
        assert sampling == "1H"

        # apply selection if given (only use subset for seasonal hourly means)
        if sel_opts is not None:
            data = data.sel(**sel_opts)

        # create unity xarray in monthly resolution with sampling point in mid of each month
        monthly = self.create_monthly_unity_array(data, time_dim) * np.nan

        # calculate for each hour in each month a separate mean value
        seasonal_hourly_means = self._compute_hourly_mean_per_month(data, time_dim, as_anomaly)

        # create seasonal cycles of these hourly averages
        seasonal_coll = []
        for hour in data.groupby(f"{time_dim}.hour").groups.keys():
            mean_single_hour = self._create_seasonal_cycle_of_single_hour_mean(monthly, seasonal_hourly_means, hour,
                                                                               time_dim, sampling)
            seasonal_coll.append(mean_single_hour)

        # combine all cycles in a common data array
        hourly = xr.concat(seasonal_coll, time_dim).sortby(time_dim).resample({time_dim: sampling}).interpolate()
        return hourly

    @staticmethod
    def extend_apriori(data: xr.DataArray, apriori: xr.DataArray, time_dim: str, sampling: str = "1d",
                       display_name: str = None) -> xr.DataArray:
        """
        Extend time range of apriori information to span a longer period as data (or at least of equal length). This
        method may not working properly if length of apriori contains data from less then one year.

        :param data: data to get time range of which apriori should span in minimum
        :param apriori: data that is adjusted. It is assumed that this data varies in the course of the year but is same
            for the same day in different years. Otherwise this method will introduce some unintended artefacts in the
            apriori data.
        :param time_dim: name of temporal dimension
        :param sampling: sampling of data (e.g. "1m", "1d", default "1d")
        :param display_name: name to use for logging message (default None)
        :returns: array which adjusted temporal coverage derived from apriori
        """
        dates = data.coords[time_dim].values
        td_type = {"1d": "D", "1H": "h"}.get(sampling)

        # apriori starts after data
        if dates[0] < apriori.coords[time_dim].values[0]:
            logging.debug(f"{display_name}: apriori starts after data")

            # add difference in full years
            date_diff = abs(dates[0] - apriori.coords[time_dim].values[0]).astype("timedelta64[D]")
            extend_range = np.ceil(date_diff / (np.timedelta64(1, "D") * 365)).astype(int) * 365
            factor = 1 if td_type == "D" else 24

            # get fill data range
            start = apriori.coords[time_dim][0].values.astype("datetime64[%s]" % td_type)
            end = apriori.coords[time_dim][0].values.astype("datetime64[%s]" % td_type) + np.timedelta64(
                366 * factor + 1, td_type)

            # fill year by year
            for i in range(365, extend_range + 365, 365):
                apriori_tmp = apriori.sel({time_dim: slice(start, end)})  # hint: slice includes end date
                new_time_axis = apriori_tmp.coords[time_dim] - np.timedelta64(i * factor, td_type)
                apriori_tmp.coords[time_dim] = new_time_axis
                apriori = apriori.combine_first(apriori_tmp)

        # apriori ends before data
        if dates[-1] + np.timedelta64(365, "D") > apriori.coords[time_dim].values[-1]:
            logging.debug(f"{display_name}: apriori ends before data")

            # add difference in full years + 1 year (because apriori is used as future estimate)
            date_diff = abs(dates[-1] - apriori.coords[time_dim].values[-1]).astype("timedelta64[D]")
            extend_range = np.ceil(date_diff / (np.timedelta64(1, "D") * 365)).astype(int) * 365 + 365
            factor = 1 if td_type == "D" else 24

            # get fill data range
            start = apriori.coords[time_dim][-1].values.astype("datetime64[%s]" % td_type) - np.timedelta64(
                366 * factor + 1, td_type)
            end = apriori.coords[time_dim][-1].values.astype("datetime64[%s]" % td_type)

            # fill year by year
            for i in range(365, extend_range + 365, 365):
                apriori_tmp = apriori.sel({time_dim: slice(start, end)})  # hint: slice includes end date
                new_time_axis = apriori_tmp.coords[time_dim] + np.timedelta64(i * factor, td_type)
                apriori_tmp.coords[time_dim] = new_time_axis
                apriori = apriori.combine_first(apriori_tmp)

        return apriori

    def combine_observation_and_apriori(self, data: xr.DataArray, apriori: xr.DataArray, time_dim: str, new_dim: str,
                                        extend_length_history: int, extend_length_future: int,
                                        extend_length_separator: int = 0) -> xr.DataArray:
        """
        Combine historical data / observations ("data") and climatological statistics ("apriori"). Historical data are
        used on interval [t0 - extend_length_history, t0] and apriori is used on [t0 + 1, t0 + extend_length_future]. If
        indicated by the extend_length_seperator, it is possible to shift end of history interval and start of apriori
        interval by given number of time steps.

        :param data: historical data for past values, must contain dimensions time_dim and var_dim and might also have
            a new_dim dimension
        :param apriori: climatological estimate for future values, must contain dimensions time_dim and var_dim, but
            can also have dimension new_dim
        :param time_dim: name of temporal dimension
        :param new_dim: name of new dim on which data is combined along
        :param extend_length_history: number of time steps to use from data
        :param extend_length_future: number of time steps to use from apriori (minus 1)
        :param extend_length_separator: position of last history value to use (default 0), this position indicates the
            last value that is used from data (followed by values from apriori). In other words, end of history
            interval and start of apriori interval are shifted by this value from t0 (positive or negative).
        :returns: combined data array
        """
        # prepare historical data / observation
        ext_sep = min(extend_length_separator, extend_length_future)
        if new_dim not in data.coords:
            history = self._shift_data(data, range(int(-extend_length_history), ext_sep + 1),
                                       time_dim, new_dim)
        else:
            history = data.sel({new_dim: slice(int(-extend_length_history), ext_sep)})

        if extend_length_future > ext_sep + 1:
            # prepare climatological statistics
            if new_dim not in apriori.coords:
                future = self._shift_data(apriori, range(ext_sep + 1,
                                                         extend_length_future + 1),
                                          time_dim, new_dim)
            else:
                future = apriori.sel({new_dim: slice(ext_sep + 1,
                                                     extend_length_future)})
            # combine historical data [t0-length,t0+sep] and climatological statistics [t0+sep+1,t0+length]
            filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
            return filter_input_data
        else:
            return history

    def create_visualization(self, filtered, data, filter_input_data, plot_dates, time_dim, new_dim, sampling,
                             extend_length_history, extend_length_future, minimum_length, h,
                             variable_name, extend_length_opts=None, offset=None):  # pragma: no cover
        plot_data = []
        offset = 0 if offset is None else offset
        extend_length_opts = 0 if extend_length_opts is None else extend_length_opts
        for t0 in set(plot_dates).intersection(filtered.coords[time_dim].values):
            try:
                td_type = {"1d": "D", "1H": "h"}.get(sampling)
                t_minus = t0 + np.timedelta64(int(-extend_length_history), td_type)
                t_plus = t0 + np.timedelta64(int(extend_length_future + 1), td_type)
                if new_dim not in data.coords:
                    tmp_filter_data = self._shift_data(data.sel({time_dim: slice(t_minus, t_plus)}),
                                                       range(int(-extend_length_history),
                                                             int(extend_length_future + 1)),
                                                       time_dim,
                                                       new_dim).sel({time_dim: t0})
                else:
                    tmp_filter_data = None
                valid_start = int(filtered[new_dim].min()) + int((len(h) + 1) / 2)
                valid_end = min(extend_length_opts + offset + 1, int(filtered[new_dim].max()) - int((len(h) + 1) / 2))
                valid_range = range(valid_start, valid_end)
                plot_data.append({"t0": t0,
                                  "var": variable_name,
                                  "filter_input": filter_input_data.sel({time_dim: t0}),
                                  "filter_input_nc": tmp_filter_data,
                                  "valid_range": valid_range,
                                  "time_range": data.sel(
                                      {time_dim: slice(t_minus, t_plus - np.timedelta64(1, td_type))}).coords[
                                      time_dim].values,
                                  "h": h,
                                  "new_dim": new_dim})
            except:
                pass
        return plot_data

    @staticmethod
    def _get_year_interval(data: xr.DataArray, time_dim: str) -> Tuple[int, int]:
        """
        Get year of start and end date of given data.

        :param data: data to extract dates from
        :param time_dim: name of temporal axis
        :returns: two-element tuple with start and end
        """
        start = pd.to_datetime(data.coords[time_dim].min().values).year
        end = pd.to_datetime(data.coords[time_dim].max().values).year
        return start, end

    @staticmethod
    def _calculate_filter_coefficients(window: Union[str, tuple], order: Union[int, tuple], cutoff_high: float,
                                       fs: float) -> np.array:
        """
        Calculate filter coefficients for moving window using scipy's signal package for common filter types and local
        method firwin_kzf for Kolmogorov Zurbenko filter (kzf). The filter is a low-pass filter.

        :param window: name of the window type which is either a string with the window's name or a tuple containing the
            name but also some parameters (e.g. `("kaiser", 5)`)
        :param order: order of the filter to create as int or parameters m and k of kzf
        :param cutoff_high: cutoff frequency to use for low-pass filter in frequency of fs
        :param fs: sampling frequency of time series
        """
        if window == "kzf":
            h = firwin_kzf(*order)
        else:
            h = signal.firwin(order, cutoff_high, pass_zero="lowpass", fs=fs, window=window)
        return h

    @staticmethod
    def _trim_data_to_minimum_length(data: xr.DataArray, extend_length_history: int, dim: str,
                                     extend_length_future: int = 0) -> xr.DataArray:
        """
        Trim data along given axis between either -minimum_length (if given) or -extend_length_history and
        extend_length_opts (which is default set to 0).

        :param data: data to trim
        :param extend_length_history: start number for trim range (transformed to negative), only used if parameter
            minimum_length is not provided
        :param dim: dim to apply trim on
        :param extend_length_future: number to use in "future"
        :returns: trimmed data
        """
        return data.sel({dim: slice(-extend_length_history, extend_length_future)}, drop=True)

    @staticmethod
    def _create_full_filter_result_array(template_array: xr.DataArray, result_array: xr.DataArray, new_dim: str,
                                         display_name: str = None) -> xr.DataArray:
        """
        Create result filter array with same shape line given template data (should be the original input data before
        filtering the data). All gaps are filled by nans.

        :param template_array: this array is used as template for shape and ordering of dims
        :param result_array: array with data that are filled into template
        :param new_dim: new dimension which is shifted/appended to/at the end (if present or not)
        :param display_name: string that is attached to logging (default None)
        """
        logging.debug(f"{display_name}: create res_full")
        new_coords = {**{k: template_array.coords[k].values for k in template_array.coords if k != new_dim},
                      new_dim: result_array.coords[new_dim]}
        dims = [*template_array.dims, new_dim] if new_dim not in template_array.dims else template_array.dims
        result_array = result_array.transpose(*dims)
        return result_array.broadcast_like(xr.DataArray(dims=dims, coords=new_coords))

    @TimeTrackingWrapper
    def clim_filter(self, data, fs, cutoff_high, order, apriori=None, sel_opts=None,
                    sampling="1d", time_dim="datetime", var_dim="variables", window: Union[str, Tuple] = "hamming",
                    minimum_length=0, next_order=0, new_dim="window", plot_dates=None, display_name=None,
                    extend_opts: int = 0, offset: int = 0):

        logging.debug(f"{display_name}: extend apriori")

        # calculate apriori information from data if not given and extend its range if not sufficient long enough
        if apriori is None:
            apriori = self.create_monthly_mean(data, time_dim, sel_opts=sel_opts, sampling=sampling)
        apriori = apriori.astype(data.dtype)
        apriori = self.extend_apriori(data, apriori, time_dim, sampling, display_name=display_name)

        # calculate FIR filter coefficients
        h = self._calculate_filter_coefficients(window, order, cutoff_high, fs)
        length = len(h)

        # set data extend that is required for filtering
        extend_length_history = minimum_length + int((next_order + 1) / 2) + int((length + 1) / 2) - offset
        extend_length_future = offset + int((next_order + 1) / 2) + int((length + 1) / 2)

        # collect some data for visualization
        plot_pos = np.array([0.25, 1.5, 2.75, 4]) * 365 * fs
        if plot_dates is None:
            plot_dates = [data.isel({time_dim: int(pos)}).coords[time_dim].values for pos in plot_pos if
                          pos < len(data.coords[time_dim])]
        plot_data = []

        coll = []
        coll_input = []

        for var in reversed(data.coords[var_dim].values):
            logging.info(f"{display_name} ({var}): sel data")

            _start, _end = self._get_year_interval(data, time_dim)
            filt_coll = []
            filt_input_coll = []
            for _year in range(_start, _end + 1):
                logging.debug(f"{display_name} ({var}): year={_year}")

                # select observations and apriori data
                time_slice = self._create_time_range_extend(
                    _year, sampling, max(extend_length_history, extend_length_future))
                d = data.sel({var_dim: [var], time_dim: time_slice})
                a = apriori.sel({var_dim: [var], time_dim: time_slice})
                if len(d.coords[time_dim]) == 0:  # no data at all for this year
                    continue

                # combine historical data / observation [t0-length,t0] and climatological statistics [t0+1,t0+length]
                filter_input_data = self.combine_observation_and_apriori(d, a, time_dim, new_dim, extend_length_history,
                    extend_length_future, extend_length_separator=extend_opts)

                # select only data for current year
                try:
                    filter_input_data = filter_input_data.sel({time_dim: str(_year)})
                except KeyError:  # no valid data for this year
                    continue
                if len(filter_input_data.coords[time_dim]) == 0:  # no valid data for this year
                    continue

                # apply filter
                logging.debug(f"{display_name} ({var}): start filter convolve")
                with TimeTracking(name=f"{display_name} ({var}): filter convolve", logging_level=logging.DEBUG):
                    filt = xr.apply_ufunc(fir_filter_convolve, filter_input_data,
                                          input_core_dims=[[new_dim]], output_core_dims=[[new_dim]],
                                          vectorize=True, kwargs={"h": h}, output_dtypes=[d.dtype])

                # trim data if required
                valid_range_end = int(filt.coords[new_dim].max() - (length + 1) / 2) + 1
                ext_len = min(extend_length_future, valid_range_end)
                trimmed = self._trim_data_to_minimum_length(filt, extend_length_history, new_dim,
                                                            extend_length_future=ext_len)
                filt_coll.append(trimmed)
                trimmed = self._trim_data_to_minimum_length(filter_input_data, extend_length_history, new_dim,
                                                            extend_length_future=ext_len)
                filt_input_coll.append(trimmed)

                # visualization
                plot_data.extend(self.create_visualization(filt, d, filter_input_data, plot_dates, time_dim, new_dim,
                                                           sampling, extend_length_history, extend_length_future,
                                                           minimum_length, h, var, extend_opts, offset))

            # collect all filter results
            coll.append(xr.concat(filt_coll, time_dim))
            coll_input.append(xr.concat(filt_input_coll, time_dim))
            gc.collect()

        # concat all variables
        logging.debug(f"{display_name}: concat all variables")
        res = xr.concat(coll, var_dim)
        res_input = xr.concat(coll_input, var_dim)

        # create result array with same shape like input data, gaps are filled by nans
        res_full = self._create_full_filter_result_array(data, res, new_dim, display_name)
        res_input_full = self._create_full_filter_result_array(data, res_input, new_dim, display_name)
        return res_full, res_input_full, h, apriori, plot_data

    @staticmethod
    def _create_time_range_extend(year: int, sampling: str, extend_length: int) -> slice:
        """
        Create a slice object for given year plus extend_length in sampling resolution.

        :param year: year to create time range for
        :param sampling: sampling of time range
        :param extend_length: number of time steps to extend out of given year
        :returns: slice object with time range
        """
        td_type = {"1d": "D", "1H": "h"}.get(sampling)
        delta = np.timedelta64(extend_length + 1, td_type)
        start = np.datetime64(f"{year}-01-01") - delta
        end = np.datetime64(f"{year}-12-31") + delta
        return slice(start, end)

    @staticmethod
    def _create_tmp_dimension(data: xr.DataArray) -> str:
        """
        Create a tmp dimension with name 'window' preferably. If name is already part of one dimensions, tmp dimension
        name is multiplied by itself until not present in dims. Method will raise ValueError after 10 tries.

        :param data: data array to create a new tmp dimension for with unique name
        :returns: valid name for a tmp dimension (preferably 'window')
        """
        new_dim = "window"
        count = 0
        while new_dim in data.dims:
            new_dim += new_dim
            count += 1
            if count > 10:
                raise ValueError("Could not create new dimension.")
        return new_dim

    def _shift_data(self, data: xr.DataArray, index_value: range, time_dim: str, new_dim: str) -> xr.DataArray:
        """
        Shift data multiple times to create history or future along dimension new_dim for each time step.

        :param data: data set to shift
        :param index_value: range of integers to span history and/or future
        :param time_dim: name of temporal dimension that should be shifted
        :param new_dim: name of dimension create by data shift
        :return: shifted data
        """
        coll = []
        for i in index_value:
            coll.append(data.shift({time_dim: -i}))
        new_ind = self.create_index_array(new_dim, index_value)
        return xr.concat(coll, dim=new_ind)

    @staticmethod
    def create_index_array(index_name: str, index_value: range):
        """
        Create index array from a range object to use as index of a data array.

        :param index_name: name of the index dimension
        :param index_value: range of values to use as indexes
        :returns: index array for given range of values
        """
        ind = pd.DataFrame({'val': index_value}, index=index_value)
        tmp_dim = index_name + "tmp"
        res = xr.Dataset.from_dataframe(ind).to_array(tmp_dim).rename({'index': index_name})
        res = res.squeeze(dim=tmp_dim, drop=True)
        res.name = index_name
        return res

    @property
    def apriori_data(self):
        return self._apriori_list

    @property
    def initial_apriori_data(self):
        return self.apriori_data[0]


def fir_filter(data, fs, order=5, cutoff_low=None, cutoff_high=None, window="hamming", dim="variables", h=None,
               causal=True, padlen=None):
    """Expects xarray."""
    if h is None:
        cutoff = []
        if cutoff_low is not None:
            cutoff += [cutoff_low]
        if cutoff_high is not None:
            cutoff += [cutoff_high]
        if len(cutoff) == 2:
            filter_type = "bandpass"
        elif len(cutoff) == 1 and cutoff_low is not None:
            filter_type = "highpass"
        elif len(cutoff) == 1 and cutoff_high is not None:
            filter_type = "lowpass"
        else:
            raise ValueError("Please provide either cutoff_low or cutoff_high.")
        h = signal.firwin(order, cutoff, pass_zero=filter_type, fs=fs, window=window)
    filtered = xr.ones_like(data)
    for var in data.coords[dim]:
        d = data.sel({dim: var}).values.flatten()
        if causal:
            y = signal.lfilter(h, 1., d)
        else:
            padlen = padlen if padlen is not None else 3 * len(h)
            y = signal.filtfilt(h, 1., d, padlen=padlen)
        filtered.loc[{dim: var}] = y
    return filtered, h


def fir_filter_convolve(data, h):
    return signal.convolve(data, h, mode='same', method="direct") / sum(h)


class KolmogorovZurbenkoBaseClass:

    def __init__(self, df, wl, itr, is_child=False, filter_dim="window"):
        """
        It create the variables associate with the Kolmogorov-Zurbenko-filter.

        Args:
            df(pd.DataFrame, None): time series of a variable
            wl(list of int): window length
            itr(list of int): number of iteration
        """
        self.df = df
        self.filter_dim = filter_dim
        self.wl = to_list(wl)
        self.itr = to_list(itr)
        if abs(len(self.wl) - len(self.itr)) > 0:
            raise ValueError("Length of lists for wl and itr must agree!")
        self._isChild = is_child
        self.child = self.set_child()
        self.type = type(self).__name__

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoBaseClass(None, self.wl[1:], self.itr[1:], True, self.filter_dim)
        else:
            return None

    def kz_filter(self, df, m, k):
        pass

    def spectral_calc(self):
        df_start = self.df
        kz = self.kz_filter(df_start, self.wl[0], self.itr[0])
        filtered = self.subtract(df_start, kz)
        # case I: no child avail -> return kz and remaining
        if self.child is None:
            return [kz, filtered]
        # case II: has child -> return current kz and all child results
        else:
            self.child.df = filtered
            kz_next = self.child.spectral_calc()
            return [kz] + kz_next

    @staticmethod
    def subtract(minuend, subtrahend):
        try:  # pandas implementation
            return minuend.sub(subtrahend, axis=0)
        except AttributeError:  # general implementation
            return minuend - subtrahend

    def run(self):
        return self.spectral_calc()

    def transfer_function(self):
        m = self.wl[0]
        k = self.itr[0]
        omega = np.linspace(0.00001, 0.15, 5000)
        return omega, (np.sin(m * np.pi * omega) / (m * np.sin(np.pi * omega))) ** (2 * k)

    def omega_null(self, alpha=0.5):
        a = np.sqrt(6) / np.pi
        b = 1 / (2 * np.array(self.itr))
        c = 1 - alpha ** b
        d = np.array(self.wl) ** 2 - alpha ** b
        return a * np.sqrt(c / d)

    def period_null(self, alpha=0.5):
        return 1. / self.omega_null(alpha)

    def period_null_days(self, alpha=0.5):
        return self.period_null(alpha) / 24.

    def plot_transfer_function(self, fig=None, name=None):
        if fig is None:
            fig = plt.figure()
        omega, transfer_function = self.transfer_function()
        if self.child is not None:
            transfer_function_child = self.child.plot_transfer_function(fig)
        else:
            transfer_function_child = transfer_function * 0
        plt.semilogx(omega, transfer_function - transfer_function_child,
                     label="m={:3.0f}, k={:3.0f}, T={:6.2f}d".format(self.wl[0],
                                                                     self.itr[0],
                                                                     self.period_null_days()))
        plt.axvline(x=self.omega_null())
        if not self._isChild:
            locs, labels = plt.xticks()
            plt.xticks(locs, np.round(1. / (locs * 24), 1))
            plt.xlim([0.00001, 0.15])
            plt.legend()
            if name is None:
                plt.show()
            else:
                plt.savefig(name)
        else:
            return transfer_function


class KolmogorovZurbenkoFilterMovingWindow(KolmogorovZurbenkoBaseClass):

    def __init__(self, df, wl: Union[list, int], itr: Union[list, int], is_child=False, filter_dim="window",
                 method="mean", percentile=0.5):
        """
        It create the variables associate with the KolmogorovZurbenkoFilterMovingWindow class.

        Args:
            df(pd.DataFrame, xr.DataArray): time series of a variable
            wl: window length
            itr: number of iteration
        """
        self.valid_methods = ["mean", "percentile", "median", "max", "min"]
        if method not in self.valid_methods:
            raise ValueError("Method '{}' is not supported. Please select from [{}].".format(
                method, ", ".join(self.valid_methods)))
        else:
            self.method = method
            if percentile > 1 or percentile < 0:
                raise ValueError("Percentile must be in range [0, 1]. Given was {}!".format(percentile))
            else:
                self.percentile = percentile
        super().__init__(df, wl, itr, is_child, filter_dim)

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoFilterMovingWindow(self.df, self.wl[1:], self.itr[1:], is_child=True,
                                                        filter_dim=self.filter_dim, method=self.method,
                                                        percentile=self.percentile)
        else:
            return None

    @TimeTrackingWrapper
    def kz_filter_new(self, df, wl, itr):
        """
        It passes the low frequency time series.

        If filter method is from mean, max, min this method will call construct and rechunk before the actual
        calculation to improve performance. If filter method is either median or percentile this approach is not
        applicable and depending on the data and window size, this method can become slow.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            for i in np.arange(0, itr):
                print(i)
                rolling = df_itr.chunk().rolling(**kwargs)
                if self.method not in ["percentile", "median"]:
                    rolling = rolling.construct("construct").chunk("auto")
                if self.method == "median":
                    df_mv_avg_tmp = rolling.median()
                elif self.method == "percentile":
                    df_mv_avg_tmp = rolling.quantile(self.percentile)
                elif self.method == "max":
                    df_mv_avg_tmp = rolling.max("construct")
                elif self.method == "min":
                    df_mv_avg_tmp = rolling.min("construct")
                else:
                    df_mv_avg_tmp = rolling.mean("construct")
                df_itr = df_mv_avg_tmp.compute()
                del df_mv_avg_tmp, rolling
                gc.collect()
            return df_itr
        except ValueError:
            raise ValueError

    @TimeTrackingWrapper
    def kz_filter(self, df, wl, itr):
        """
        It passes the low frequency time series.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        import warnings
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            iter_vars = df_itr.coords["variables"].values
            for var in iter_vars:
                df_itr_var = df_itr.sel(variables=[var])
                for _ in np.arange(0, itr):
                    df_itr_var = df_itr_var.chunk()
                    rolling = df_itr_var.rolling(**kwargs)
                    if self.method == "median":
                        df_mv_avg_tmp = rolling.median()
                    elif self.method == "percentile":
                        df_mv_avg_tmp = rolling.quantile(self.percentile)
                    elif self.method == "max":
                        df_mv_avg_tmp = rolling.max()
                    elif self.method == "min":
                        df_mv_avg_tmp = rolling.min()
                    else:
                        df_mv_avg_tmp = rolling.mean()
                    df_itr_var = df_mv_avg_tmp.compute()
                df_itr.loc[{"variables": [var]}] = df_itr_var
            return df_itr
        except ValueError:
            raise ValueError


def firwin_kzf(m: int, k: int) -> np.array:
    """Calculate weights of window for Kolmogorov Zurbenko filter."""
    m, k = int(m), int(k)
    coef = np.ones(m)
    for i in range(1, k):
        t = np.zeros((m, m + i * (m - 1)))
        for km in range(m):
            t[km, km:km + coef.size] = coef
        coef = np.sum(t, axis=0)
    return coef / (m ** k)


def omega_null_kzf(m: int, k: int, alpha: float = 0.5) -> float:
    a = np.sqrt(6) / np.pi
    b = 1 / (2 * np.array(k))
    c = 1 - alpha ** b
    d = np.array(m) ** 2 - alpha ** b
    return a * np.sqrt(c / d)


def filter_width_kzf(m: int, k: int) -> int:
    """Returns window width of the Kolmorogov Zurbenko filter."""
    return k * (m - 1) + 1
