__author__ = "Lukas Leufen"
__date__ = '2021-02-'

from functools import reduce, partial

from mlair.model_modules import AbstractModelClass
from mlair.helpers import select_from_dict
from mlair.model_modules.loss import var_loss, custom_loss
from mlair.model_modules.advanced_paddings import PadUtils, Padding2D, SymmetricPadding2D

import tensorflow.keras as keras


class CNN(AbstractModelClass):  # pragma: no cover

    _activation = {"relu": keras.layers.ReLU, "tanh": partial(keras.layers.Activation, "tanh"),
                   "sigmoid": partial(keras.layers.Activation, "sigmoid"),
                   "linear": partial(keras.layers.Activation, "linear"),
                   "selu": partial(keras.layers.Activation, "selu"),
                   "prelu": partial(keras.layers.PReLU, alpha_initializer=keras.initializers.constant(value=0.25))}
    _initializer = {"tanh": "glorot_uniform", "sigmoid": "glorot_uniform", "linear": "glorot_uniform",
                    "relu": keras.initializers.he_normal(), "selu": keras.initializers.lecun_normal(),
                    "prelu": keras.initializers.he_normal()}
    _optimizer = {"adam": keras.optimizers.Adam, "sgd": keras.optimizers.SGD}
    _regularizer = {"l1": keras.regularizers.l1, "l2": keras.regularizers.l2, "l1_l2": keras.regularizers.l1_l2}
    _requirements = ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad", "momentum", "nesterov", "l1", "l2"]
    _dropout = {"selu": keras.layers.AlphaDropout}

    def __init__(self, input_shape: list, output_shape: list, activation="relu", activation_output="linear",
                 optimizer="adam", regularizer=None, kernel_size=1, dropout=None, **kwargs):

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.activation = self._set_activation(activation)
        self.activation_name = activation
        self.activation_output = self._set_activation(activation_output)
        self.activation_output_name = activation_output
        self.kernel_initializer = self._initializer.get(activation, "glorot_uniform")
        self.kernel_regularizer = self._set_regularizer(regularizer, **kwargs)
        self.kernel_size = kernel_size
        self.optimizer = self._set_optimizer(optimizer, **kwargs)
        self.dropout, self.dropout_rate = self._set_dropout(activation, dropout)

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=custom_loss([keras.losses.mean_squared_error, var_loss]), var_loss=var_loss)

    def _set_activation(self, activation):
        try:
            return self._activation.get(activation.lower())
        except KeyError:
            raise AttributeError(f"Given activation {activation} is not supported in this model class.")

    def _set_optimizer(self, optimizer, **kwargs):
        try:
            opt_name = optimizer.lower()
            opt = self._optimizer.get(opt_name)
            opt_kwargs = {}
            if opt_name == "adam":
                opt_kwargs = select_from_dict(kwargs, ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad"])
            elif opt_name == "sgd":
                opt_kwargs = select_from_dict(kwargs, ["lr", "momentum", "decay", "nesterov"])
            return opt(**opt_kwargs)
        except KeyError:
            raise AttributeError(f"Given optimizer {optimizer} is not supported in this model class.")

    def _set_regularizer(self, regularizer, **kwargs):
        if regularizer is None or (isinstance(regularizer, str) and regularizer.lower() == "none"):
            return None
        try:
            reg_name = regularizer.lower()
            reg = self._regularizer.get(reg_name)
            reg_kwargs = {}
            if reg_name in ["l1", "l2"]:
                reg_kwargs = select_from_dict(kwargs, reg_name, remove_none=True)
                if reg_name in reg_kwargs:
                    reg_kwargs["l"] = reg_kwargs.pop(reg_name)
            elif reg_name == "l1_l2":
                reg_kwargs = select_from_dict(kwargs, ["l1", "l2"], remove_none=True)
            return reg(**reg_kwargs)
        except KeyError:
            raise AttributeError(f"Given regularizer {regularizer} is not supported in this model class.")

    def _set_dropout(self, activation, dropout_rate):
        if dropout_rate is None:
            return None, None
        assert 0 <= dropout_rate < 1
        return self._dropout.get(activation, keras.layers.Dropout), dropout_rate

    def set_model(self):
        """
        Build the model.
        """
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(filters=16, kernel_size=(73, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_input)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(filters=32, kernel_size=(49, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        if self.dropout is not None:
            x_in = self.dropout(self.dropout_rate)(x_in)
        x_in = keras.layers.MaxPooling2D((25, 1), strides=(1, 1), padding='valid')(x_in)
        x_in = keras.layers.Conv2D(filters=64, kernel_size=(13, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        if self.dropout is not None:
            x_in = self.dropout(self.dropout_rate)(x_in)
        x_in = keras.layers.Flatten()(x_in)
        x_in = keras.layers.Dense(128, kernel_initializer=self.kernel_initializer,
                                  kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(32, kernel_initializer=self.kernel_initializer,
                                  kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])

    def set_compile_options(self):
        self.compile_options = {"loss": [custom_loss([keras.losses.mean_squared_error, var_loss])],
                                "metrics": ["mse", "mae", var_loss]}
