"""
Module for neural models to use during experiment.

To work properly, each customised model needs to inherit from AbstractModelClass and needs an implementation of the
set_model method.

In this module, you can find some exemplary model classes that have been build and were running in a experiment.

* `MyLittleModel`: small model implementation with a single 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time).
* `MyBranchedModel`: a model with single  1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), it has three
  output branches from different layers of the model.
* `MyTowerModel`: a more complex model with inception blocks (called towers)
* `MyPaperModel`: A model used for the publication: <Add Publication Title / Citation>

In addition, a short introduction how to create your own model is given hereinafter.

How to create a customised model?
#################################

* Create a new class:

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def __init__(self, input_shape: list, output_shape: list):

                super().__init__(input_shape[0], output_shape[0])

                # settings
                self.dropout_rate = 0.1
                self.activation = keras.layers.PReLU

                # apply to model
                self.set_model()
                self.set_compile_options()
                self.set_custom_objects(loss=self.compile_options['loss'])

* Make sure to add the `super().__init__()` and at least `set_model()` and `set_compile_options()` to your custom init
  method.
* If you have custom objects in your model, that are not part of keras, you need to add them to custom objects. To do
  this, call `set_custom_objects` with arbitrarily kwargs. In the shown example, the loss has been added, because it
  wasn't a standard loss. Apart from this, we always encourage you to add the loss as custom object, to prevent
  potential errors when loading an already created model instead of training a new one.
* Build your model inside `set_model()`, e.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                x_input = keras.layers.Input(shape=self._input_shape)
                x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
                x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
                x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
                x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
                x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
                x_in = self.activation()(x_in)
                x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
                out_main = self.activation()(x_in)
                self.model = keras.Model(inputs=x_input, outputs=[out_main])

* Your are free, how to design your model. Just make sure to save it in the class attribute model.
* Additionally, set your custom compile options including the loss.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_compile_options(self):
                self.initial_lr = 1e-2
                self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
                self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
                                                                                       drop=.94,
                                                                                       epochs_drop=10)
                self.loss = keras.losses.mean_squared_error
                self.compile_options = {"metrics": ["mse", "mae"]}

* If you have a branched model with multiple outputs, you need either set only a single loss for all branch outputs or
  to provide the same number of loss functions considering the right order. E.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                ...
                self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

            def set_compile_options(self):
                self.loss = [keras.losses.mean_absolute_error] +  # for out_minor_1
                            [keras.losses.mean_squared_error] +   # for out_minor_2
                            [keras.losses.mean_squared_error]     # for out_main


How to access my customised model?
##################################

If the customised model is created, you can easily access the model with

>>> MyCustomisedModel().model
<your custom model>

The loss is accessible via

>>> MyCustomisedModel().loss
<your custom loss>

You can treat the instance of your model as instance but also as the model itself. If you call a method, that refers to
the model instead of the model instance, you can directly apply the command on the instance instead of adding the model
parameter call.

>>> MyCustomisedModel().model.compile(**kwargs) == MyCustomisedModel().compile(**kwargs)
True

"""

import mlair.model_modules.keras_extensions

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2020-05-12'

import tensorflow.keras as keras

from mlair.model_modules import AbstractModelClass
from mlair.model_modules.inception_model import InceptionModelBase
from mlair.model_modules.flatten import flatten_tail
from mlair.model_modules.advanced_paddings import PadUtils, Padding2D, SymmetricPadding2D
from mlair.model_modules.loss import l_p_loss


class MyLittleModelHourly(AbstractModelClass):
    """
    A customised model with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param shape_inputs: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param shape_outputs: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.001)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(128, (1, 1), padding='same', name='{}_Conv_1x1_128'.format("major"))(x_input)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(64, (1, 1), padding='same', name='{}_Conv_1x1_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(128, name='{}_Dense_128'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyBranchedModel(AbstractModelClass):
    """
    A customised model

    with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.1)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
        x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_1 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_1"))(x_in)
        out_minor_1 = self.activation(name="minor_1")(out_minor_1)
        x_in = keras.layers.Dense(32, name='{}_Dense_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_2 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_2"))(x_in)
        out_minor_2 = self.activation(name="minor_2")(out_minor_2)
        x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation(name="main")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.compile_options = {"loss": [keras.losses.mean_absolute_error] + [keras.losses.mean_squared_error] + [
            keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyTowerModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 1e-2
        self.regularizer = keras.regularizers.l2(0.1)
        self.initial_lr = 1e-2
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """
        activation = self.activation
        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (3, 1), 'activation': activation},
            'tower_2': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (5, 1), 'activation': activation},
            'tower_3': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (1, 1), 'activation': activation},
        }

        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 8 * 2, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (1, 1),
                        'activation': activation},
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict3 = {'tower_1': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                                           'activation': activation},
                               'tower_2': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
                                           'activation': activation},
                               'tower_3': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
                                           'activation': activation},
                               }

        pool_settings_dict3 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        X_in = inception_model.inception_block(X_input, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict3, pool_settings_dict3,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)
        #############################################

        # out_main = flatten_tail(X_in, 'Main', activation=activation, bound_weight=True, dropout_rate=self.dropout_rate,
        #                         reduction_filter=64, inner_neurons=64, output_neurons=self.window_lead_time)

        out_main = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64,
                                name='Main', bound_weight=True, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse"]}


class IntelliO3TsArchitecture(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        super().__init__(input_shape[0], output_shape[0])

        from mlair.model_modules.keras_extensions import LearningRateDecay

        # settings
        self.dropout_rate = .35
        self.regularizer = keras.regularizers.l2(0.01)
        self.initial_lr = 1e-4
        self.lr_decay = LearningRateDecay(base_lr=self.initial_lr, drop=.94, epochs_drop=10)
        self.activation = keras.layers.ELU
        self.padding = "SymPad2D"

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"],
                                SymmetricPadding2D=SymmetricPadding2D,
                                LearningRateDecay=LearningRateDecay)

    def set_model(self):
        """
        Build the model.

        :param activation: activation function
        :param window_history_size: number of historical time steps included in the input data
        :param channels: number of variables used in input data
        :param dropout_rate: dropout rate used in the model [0, 1)
        :param window_lead_time: number of time steps to forecast in the output layer
        :return: built keras model
        """
        activation = self.activation
        first_kernel = (3, 1)
        first_filters = 16

        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (1, 1),
                        'activation': activation}
        }
        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
                        'activation': activation}
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        pad_size = PadUtils.get_padding_for_same(first_kernel)
        X_in = Padding2D("SymPad2D")(padding=pad_size, name="SymPad")(X_input)
        X_in = keras.layers.Conv2D(filters=first_filters,
                                   kernel_size=first_kernel,
                                   kernel_regularizer=self.regularizer,
                                   name="First_conv_{}x{}".format(first_kernel[0], first_kernel[1]))(X_in)
        X_in = self.activation(name='FirstAct')(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True,
                                               padding=self.padding)

        out_minor1 = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                  output_activation='linear', reduction_filter=32 * 2,
                                  name='minor_1', bound_weight=False, dropout_rate=self.dropout_rate,
                                  kernel_regularizer=self.regularizer
                                  )

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True, padding=self.padding)

        #############################################

        out_main = flatten_tail(X_in, inner_neurons=64 * 2, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64 * 2,
                                name='Main', bound_weight=False, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_minor1, out_main])

    def set_compile_options(self):
        self.compile_options = {"optimizer": keras.optimizers.Adam(lr=self.initial_lr, amsgrad=True),
                                "loss": [l_p_loss(4), keras.losses.mean_squared_error],
                                "metrics": ['mse'],
                                "loss_weights": [.01, .99]
                                }


class MyLSTMModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):

        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.4
        self.stateful = False
        self.initial_lr = 1e-3 # Adam Default: 1e-3
        self.amsgrad = True # Default: False

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.LSTM(16*2, return_sequences=True, name="First_LSTM", dropout=self.dropout_rate, stateful=self.stateful)(x_input)
        # x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        x_in = keras.layers.LSTM(16*2, name="Second_LSTM", dropout=self.dropout_rate, stateful=self.stateful)(x_in)
        out_main = keras.layers.Dense(self._output_shape, name='Output_Dense')(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
#         self.initial_lr = 1e-4
        #self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr, amsgrad=self.amsgrad)
#         self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
#                                                                                drop=.94,
#                                                                                epochs_drop=10)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}


class MySimpleGRU(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape[0], output_shape[0])
        # settings
        self.dropout_rate = 0.1
        self.initial_lr = 0.01
        self.clipnorm = 1

        self.n_hidden = 100

        # apply to model
        self.set_model()
        self.set_compile_options()
        # self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        x = keras.layers.GRU(self.n_hidden, dropout=self.dropout_rate, activation='elu')(input_train)
        out = keras.layers.Dense(self._output_shape)(x)
        self.model = keras.Model(inputs=input_train, outputs=[out])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr, clipnorm=self.clipnorm)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}


class MySimpleConv1D(AbstractModelClass):
    """
    Example adopted from https://www.kaggle.com/dimitreoliveira/deep-learning-for-time-series-forecasting
    """

    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape[0], output_shape[0])
        self.filters = 16
        self.kernel_size = 2
        self.activation = "elu"
        self.pool_size = 2

        self.dropout = .4
        self.kernel_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)
        self.bias_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)

        self.dense_units = 32
        self.initial_lr = 0.0001

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(SymmetricPadding2D=SymmetricPadding2D)
        # self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        x = keras.layers.Conv1D(filters=self.filters, kernel_size=self.kernel_size, activation=self.activation,
                                kernel_regularizer=self.kernel_regularizer, bias_regularizer=self.bias_regularizer
                                )(input_train)
        x = keras.layers.MaxPool1D(pool_size=self.pool_size)(x)
        #####
        x = keras.layers.Conv1D(filters=self.filters*2, kernel_size=self.kernel_size, activation=self.activation,
                                kernel_regularizer=self.kernel_regularizer, bias_regularizer=self.bias_regularizer
                                )(x)
        #x = keras.layers.MaxPool1D(pool_size=self.pool_size)(x)
        #####
        x = keras.layers.Flatten()(x)
        x = keras.layers.Dropout(self.dropout)(x)
        x = keras.layers.Dense(units=self.dense_units, activation=self.activation)(x)
        #####
        x = keras.layers.Dropout(self.dropout)(x)
        out = keras.layers.Dense(units=self._output_shape)(x)

        self.model = keras.Model(inputs=input_train, outputs=[out])

    def set_compile_options(self):

        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr,
                                              # clipnorm=self.clipnorm,
                                              )
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
        
        
class MyUnet(AbstractModelClass):
    """
    Original source code from https://www.kaggle.com/keegil/keras-u-net-starter-lb-0-277
    Modified version including LSTM layers
    """

    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape[0], output_shape[0])
        self.first_filter_size = 16*2#self._input_shape[-1]  # 16
        self.lstm_units = 64 * 2
        self.kernel_size = (3, 1)  # (3,1)
        self.activation = "elu"
        self.pool_size = (2, 1)

        # self.dropout = .25
        self.kernel_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)
        self.bias_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)

        self.kernel_initializer = 'he_normal'

        self.dense_units = 32*2
        self.initial_lr = 0.001

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(SymmetricPadding2D=SymmetricPadding2D)

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        pad_size = PadUtils.get_padding_for_same(self.kernel_size)

        c1 = Padding2D("SymPad2D")(padding=pad_size)(input_train)
        c1 = keras.layers.Conv2D(self.first_filter_size, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c1)
        c1 = keras.layers.Dropout(0.1)(c1)
        c1 = Padding2D("SymPad2D")(padding=pad_size)(c1)
        c1 = keras.layers.Conv2D(self.first_filter_size, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c1',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c1)
        p1 = c1
        # p1 = keras.layers.MaxPooling2D(self.pool_size)(c1)

        c2 = Padding2D("SymPad2D")(padding=pad_size)(p1)
        c2 = keras.layers.Conv2D(self.first_filter_size * 2, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c2)
        c2 = keras.layers.Dropout(0.1)(c2)
        c2 = Padding2D("SymPad2D")(padding=pad_size)(c2)
        c2 = keras.layers.Conv2D(self.first_filter_size * 2, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c2',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c2)
        p2 = c2
        # p2 = keras.layers.MaxPooling2D(self.pool_size)(c2)

        c3 = Padding2D("SymPad2D")(padding=pad_size)(p2)
        c3 = keras.layers.Conv2D(self.first_filter_size * 4, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c3)
        c3 = keras.layers.Dropout(0.2)(c3)
        c3 = Padding2D("SymPad2D")(padding=pad_size)(c3)
        c3 = keras.layers.Conv2D(self.first_filter_size * 4, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c3',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c3)
        # p3 = c3
        p3 = keras.layers.MaxPooling2D(self.pool_size)(c3)

        ### own LSTM Block ###
        ls1 = keras.layers.Reshape((p3.shape[1], p3.shape[-1]))(p3)
        ls1 = keras.layers.LSTM(self.lstm_units, return_sequences=True)(ls1)
        ls1 = keras.layers.LSTM(self.lstm_units, return_sequences=True)(ls1)
        c4 = keras.layers.Reshape((p3.shape[1], 1, -1))(ls1)

        ### own 2nd LSTM Block ###
        ls2 = keras.layers.Reshape((c3.shape[1], c3.shape[-1]))(c3)
        ls2 = keras.layers.LSTM(self.lstm_units, return_sequences=True)(ls2)
        ls2 = keras.layers.LSTM(self.lstm_units, return_sequences=True)(ls2)
        c4_2 = keras.layers.Reshape((c3.shape[1], 1, -1))(ls2)

        u7 = keras.layers.UpSampling2D(size=(3, 1))(c4)
        cn3 = Padding2D("SymPad2D")(padding=pad_size)(c3)
        # u7 = c4
        u7 = keras.layers.concatenate([u7, cn3], name="u7_c3")
        c7 = u7
        # c7 = Padding2D("SymPad2D")(padding=pad_size)(u7)
        c7 = keras.layers.Conv2D(self.first_filter_size * 4, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c7)
        c7 = keras.layers.concatenate([c7, c4_2], name="Concat_2nd_LSTM")
        c7 = keras.layers.Dropout(0.2)(c7)
        c7 = Padding2D("SymPad2D")(padding=pad_size)(c7)
        c7 = keras.layers.Conv2D(self.first_filter_size * 4, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c7_to_u8',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c7)

        # u8 = Padding2D("SymPad2D")(padding=pad_size)(c7)
        # u8 = keras.layers.Conv2DTranspose(32, self.pool_size, strides=self.pool_size)(u8)
        u8 = c7
        # u8 = c3
        u8 = keras.layers.concatenate([u8, c2], name="u8_c2")
        c8 = Padding2D("SymPad2D")(padding=pad_size)(u8)
        c8 = keras.layers.Conv2D(self.first_filter_size * 2, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c8)
        c8 = keras.layers.Dropout(0.1)(c8)
        c8 = Padding2D("SymPad2D")(padding=pad_size)(c8)
        c8 = keras.layers.Conv2D(self.first_filter_size * 2, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c8_to_u9',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c8)

        # u9 = Padding2D("SymPad2D")(padding=pad_size)(c8)
        # u9 = keras.layers.Conv2DTranspose(16, self.pool_size, strides=self.pool_size)(u9)
        u9 = c8
        u9 = keras.layers.concatenate([u9, c1], name="u9_c1")
        c9 = Padding2D("SymPad2D")(padding=pad_size)(u9)
        c9 = keras.layers.Conv2D(self.first_filter_size, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c9)
        c9 = keras.layers.Dropout(0.1)(c9)
        c9 = Padding2D("SymPad2D")(padding=pad_size)(c9)
        c9 = keras.layers.Conv2D(self.first_filter_size, self.kernel_size, activation=self.activation,
                                 kernel_initializer=self.kernel_initializer, name='c9',
                                 kernel_regularizer=self.kernel_regularizer,
                                 bias_regularizer=self.bias_regularizer)(c9)

        # outputs = keras.layers.Conv2D(1, (1, 1), activation='sigmoid')(c9)
        dl = keras.layers.Flatten()(c9)
        outputs = keras.layers.Dense(units=self._output_shape)(dl)

        self.model = keras.Model(inputs=[input_train], outputs=[outputs])

    def set_compile_options(self):

        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr,
                                              # clipnorm=self.clipnorm,
                                              )
        
        loss = keras.losses.mean_squared_error
        self.compile_options = {"loss": [loss], "metrics": ["mse", "mae"]}
        
class NN3s(MyUnet):
    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape, output_shape)


class MySimpleConv2D(AbstractModelClass):
    """
    Example adopted from https://www.kaggle.com/dimitreoliveira/deep-learning-for-time-series-forecasting
    """

    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape[0], output_shape[0])
        self.filters = 32
        self.lstm_units = 64
        self.kernel_size = (5,1)# (3,1)
        self.activation = "elu"
        self.pool_size = (2,1)

        self.dropout = .25
        self.kernel_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)
        self.bias_regularizer = keras.regularizers.l1_l2(l1=0.01, l2=0.01)

        self.dense_units = 32*2
        self.initial_lr = 0.001

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(SymmetricPadding2D=SymmetricPadding2D)
        # self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        
        pad_size = PadUtils.get_padding_for_same(self.kernel_size)
        x = Padding2D("SymPad2D")(padding=pad_size, name="SymPad1")(input_train)
        x = keras.layers.Conv2D(filters=self.filters, kernel_size=self.kernel_size, activation=self.activation,
                                kernel_regularizer=self.kernel_regularizer, bias_regularizer=self.bias_regularizer
                                )(x)
        #x = keras.layers.MaxPool2D(pool_size=self.pool_size)(x)
        #####
        x = Padding2D("SymPad2D")(padding=pad_size, name="SymPad2")(x)
        x = keras.layers.Conv2D(filters=self.filters*2, kernel_size=self.kernel_size, activation=self.activation,
                                kernel_regularizer=self.kernel_regularizer, bias_regularizer=self.bias_regularizer
                                )(x)
        x = keras.layers.Reshape((x.shape[1].value,-1))(x)
        x = keras.layers.LSTM(self.lstm_units)(x)

        # x = Padding2D("SymPad2D")(padding=pad_size, name="SymPad3")(x)
        #
        # x = keras.layers.Conv2D(filters=self.filters*2*2, kernel_size=self.kernel_size, activation=self.activation,
        #                         kernel_regularizer=self.kernel_regularizer, bias_regularizer=self.bias_regularizer
        #                         )(x)
        #x = keras.layers.MaxPool2D(pool_size=self.pool_size)(x)
        #####
        # x = keras.layers.Flatten()(x)
        x = keras.layers.Dropout(self.dropout)(x)
        x = keras.layers.Dense(units=self.dense_units)(x)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Activation(self.activation)(x)
        #####
        x = keras.layers.Dropout(self.dropout)(x)
        out = keras.layers.Dense(units=self._output_shape)(x)

        self.model = keras.Model(inputs=input_train, outputs=[out])

    def set_compile_options(self):

        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr,
                                              # clipnorm=self.clipnorm,
                                              )
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
        

class MySimpleLSTM(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        super().__init__(input_shape[0], output_shape[0])
        # settings
        self.dropout_rate = 0.1
        self.initial_lr = 0.01
        self.clipnorm = 1

        self.n_hidden = 64

        # apply to model
        self.set_model()
        self.set_compile_options()
        # self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        x = keras.layers.LSTM(self.n_hidden, dropout=self.dropout_rate, activation='elu')(input_train)
        out = keras.layers.Dense(self._output_shape)(x)
        self.model = keras.Model(inputs=input_train, outputs=[out])

    def set_compile_options(self):

        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr, clipnorm=self.clipnorm)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}


class MyLuongAttentionLSTMModel(AbstractModelClass):
    """
    Luong Attention LSTM from https://levelup.gitconnected.com/building-seq2seq-lstm-with-luong-attention-in-keras-for-time-series-forecasting-1ee00958decb
    """

    def __init__(self, input_shape: list, output_shape: list):

        super().__init__(input_shape[0], output_shape[0])

        ## original settings from Huangwei Wieniawska (see url above)
        self.initial_lr = 0.01 # orig: 0.01
        self.n_hidden = 100 # orig: 100
        self.activation_lstm = "elu"
        self.droput_rate = 0.2
        self.recurrent_dropout = 0.2
        self.optimizer_kwargs = {"clipnorm": 1, # orig: 1
                                }
        
        # settings
        self.droput_rate = 0.5
        self.recurrent_dropout = 0.5
         
        self.n_hidden = 16
        self.activation_lstm = "tanh"
        
        self.initial_lr = 0.001 # orig: 0.01    
        self.optimizer_kwargs = {"clipnorm": .9,
                                 "clipvalue": .99, 
                                }
        

        # apply to model
        self.set_model()
        self.set_compile_options()
        # self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        input_train = keras.layers.Input(shape=self._input_shape)
        # output_train = keras.layers.Input(shape=self._output_shape)

        encoder_stack_h, encoder_last_h, encoder_last_c = keras.layers.LSTM(
            self.n_hidden, activation=self.activation_lstm, dropout=self.droput_rate, recurrent_dropout=self.recurrent_dropout,
            return_state=True, return_sequences=True)(input_train)

        encoder_last_h = keras.layers.BatchNormalization(momentum=0.6)(encoder_last_h)
        encoder_last_c = keras.layers.BatchNormalization(momentum=0.6)(encoder_last_c)

        # decoder_input = keras.layers.RepeatVector(output_train.shape[1])(encoder_last_h)
        decoder_input = keras.layers.RepeatVector(self._output_shape)(encoder_last_h)

        decoder_stack_h = keras.layers.LSTM(self.n_hidden, activation=self.activation_lstm, 
                                            dropout=self.droput_rate, 
                                            recurrent_dropout=self.recurrent_dropout,
                                            return_state=False, return_sequences=True)(
            decoder_input, initial_state=[encoder_last_h, encoder_last_c])

        attention = keras.layers.dot([decoder_stack_h, encoder_stack_h], axes=[2, 2])
        attention = keras.layers.Activation('softmax')(attention)

        context = keras.layers.dot([attention, encoder_stack_h], axes=[2, 1])
        context = keras.layers.BatchNormalization(momentum=0.6)(context)

        decoder_combined_context = keras.layers.concatenate([context, decoder_stack_h])

        # out = keras.layers.TimeDistributed(keras.layers.Dense(output_train.shape[2]))(decoder_combined_context)
        out = keras.layers.TimeDistributed(keras.layers.Dense(1))(decoder_combined_context)

        # FK add
        out = keras.layers.Flatten()(out)

        self.model = keras.Model(inputs=input_train, outputs=[out])

    def set_compile_options(self):

        self.optimizer = keras.optimizers.Adam(lr=self.initial_lr, **self.optimizer_kwargs)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
        
        
class MyCNNModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):

        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.4
        self.batchnormalization = keras.layers.normalization.BatchNormalization
#         self.activation = keras.layers.LeakyReLU
        self.activation = keras.layers.ELU
#         self.regularizer = keras.regularizers.l2(0.001)
        
        self.initial_lr = 0.01
        

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(SymmetricPadding2D=SymmetricPadding2D)

    def set_model(self):

        first_kernel = (3, 1)
        first_filters = 16
        pad_size1 = PadUtils.get_padding_for_same(first_kernel)
        
        pool_kernel = (3,1)
        pad_size_pool = PadUtils.get_padding_for_same(pool_kernel)
        
        second_kernel = (3, 1)
        second_filters = 32
        pad_size2 = PadUtils.get_padding_for_same(second_kernel)
        
        x_input = keras.layers.Input(shape=self._input_shape)

        x_in = Padding2D("SymPad2D")(padding=pad_size1, name="SymPad1")(x_input)
        x_in = keras.layers.Conv2D(filters=first_filters,
                                   kernel_size=first_kernel,
#                                    kernel_regularizer=self.regularizer,
                                   name="First_conv_{}x{}".format(first_kernel[0], first_kernel[1]))(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        
#         x_in = Padding2D("SymPad2D")(padding=pad_size_pool, name="SymPad_pool")(x_in)
#         x_in = keras.layers.MaxPool2D(pool_kernel)(x_in)

        x_in = Padding2D("SymPad2D")(padding=pad_size2, name="SymPad2")(x_in)
        x_in = keras.layers.Conv2D(filters=second_filters,
                                   kernel_size=second_kernel,
#                                    kernel_regularizer=self.regularizer,
                                   name="Second_conv_{}x{}".format(second_kernel[0], second_kernel[1]))(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        

        x_in = keras.layers.Flatten()(x_in)

        x_in = keras.layers.Dense(32)(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        
        out_main = keras.layers.Dense(self._output_shape, name='Output_Dense')(x_in)

        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        
#         self.optimizer = keras.optimizers.Adam()
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
#         self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
#                                                                                drop=.94,
#                                                                                epochs_drop=10)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
        
class MyCNNModelSect(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):

        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.35
        self.batchnormalization = keras.layers.normalization.BatchNormalization
#         self.activation = keras.layers.LeakyReLU
        self.activation = keras.layers.ELU
#         self.regularizer = keras.regularizers.l2(0.001)
        
        self.initial_lr = 0.01
        

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(SymmetricPadding2D=SymmetricPadding2D)

    def set_model(self):

        first_kernel = (3, 1)
        first_filters = 16
        pad_size1 = PadUtils.get_padding_for_same(first_kernel)
        
        pool_kernel = (3, 1)
        pad_size_pool = PadUtils.get_padding_for_same(pool_kernel)
        
        second_kernel = (3, 1)
        second_filters = 32
        pad_size2 = PadUtils.get_padding_for_same(second_kernel)

        x_input = keras.layers.Input(shape=self._input_shape)

        x_in = Padding2D("SymPad2D")(padding=pad_size1, name="SymPad1")(x_input)
        x_in = keras.layers.Conv2D(filters=first_filters,
                                   kernel_size=first_kernel,
#                                    kernel_regularizer=self.regularizer,
                                   name="First_conv_{}x{}".format(first_kernel[0], first_kernel[1]))(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        
#         x_in = Padding2D("SymPad2D")(padding=pad_size_pool, name="SymPad_pool")(x_in)
#         x_in = keras.layers.MaxPool2D(pool_kernel)(x_in)

        x_in = Padding2D("SymPad2D")(padding=pad_size2, name="SymPad2")(x_in)
        x_in = keras.layers.Conv2D(filters=second_filters,
                                   kernel_size=second_kernel,
#                                    kernel_regularizer=self.regularizer,
                                   name="Second_conv_{}x{}".format(second_kernel[0], second_kernel[1]))(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        

        x_in = keras.layers.Flatten()(x_in)

        x_in = keras.layers.Dense(32)(x_in)
        x_in = self.batchnormalization()(x_in)
        x_in = self.activation()(x_in)
        
        x_in = keras.layers.Dropout(self.dropout_rate)(x_in)
        
        out_main = keras.layers.Dense(self._output_shape, name='Output_Dense')(x_in)

        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        
#         self.optimizer = keras.optimizers.Adam()
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
#         self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
#                                                                                drop=.94,
#                                                                                epochs_drop=10)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
